"""UNKNOWN values from :rfc:`7265`."""

from typing import ClassVar

from icalendar.compatibility import Self
from icalendar.prop.text import vText


class vUnknown(vText):
    """This is text but the VALUE parameter is unknown.

    Since :rfc:`7265`, it is important to record if values are unknown.
    For :rfc:`5545`, we could just assume TEXT.
    """

    default_value: ClassVar[str] = "UNKNOWN"

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vUnknown."""
        return [vUnknown("Some property text.")]

    from icalendar.param import VALUE


__all__ = ["vUnknown"]



"""FLOAT values from :rfc:`5545`."""

from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters


class vFloat(float):
    """Float

    Value Name:
        FLOAT

    Purpose:
        This value type is used to identify properties that contain
        a real-number value.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            float      = (["+"] / "-") 1*DIGIT ["." 1*DIGIT]

    Description:
        If the property permits, multiple "float" values are
        specified by a COMMA-separated list of values.

        Example:

        .. code-block:: text

            1000000.0000001
            1.333
            -3.14

        .. code-block:: pycon

            >>> from icalendar.prop import vFloat
            >>> float = vFloat.from_ical('1000000.0000001')
            >>> float
            1000000.0000001
            >>> float = vFloat.from_ical('1.333')
            >>> float
            1.333
            >>> float = vFloat.from_ical('+1.333')
            >>> float
            1.333
            >>> float = vFloat.from_ical('-3.14')
            >>> float
            -3.14
    """

    default_value: ClassVar[str] = "FLOAT"
    params: Parameters

    def __new__(
        cls, *args: Any, params: dict[str, Any] | None = None, **kwargs: Any
    ) -> Self:
        self = super().__new__(cls, *args, **kwargs)
        self.params = Parameters(params)
        return self

    def to_ical(self) -> bytes:
        return str(self).encode("utf-8")

    @classmethod
    def from_ical(cls, ical: str | float) -> Self:
        try:
            return cls(ical)
        except Exception as e:
            raise ValueError(f"Expected float value, got: {ical}") from e

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vFloat."""
        return [cls(3.1415)]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        return [name, self.params.to_jcal(), self.VALUE.lower(), float(self)]

    @property
    def ical_value(self) -> float:
        """Converts the FLOAT property type according to :rfc:`5545#section-3.3.7` to a Python float."""
        return float(self)

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the jCal provided is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        if jcal_property[0].upper() == "GEO":
            from icalendar.prop import vGeo

            return vGeo.from_jcal(jcal_property)
        JCalParsingError.validate_value_type(jcal_property[3], float, cls, 3)
        return cls(
            jcal_property[3],
            params=Parameters.from_jcal_property(jcal_property),
        )


__all__ = ["vFloat"]



"""URI values from :rfc:`5545`."""

from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.parser_tools import DEFAULT_ENCODING, to_unicode


class vUri(str):
    """URI

    Value Name:
        URI

    Purpose:
        This value type is used to identify values that contain a
        uniform resource identifier (URI) type of reference to the
        property value.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            uri = scheme ":" hier-part [ "?" query ] [ "#" fragment ]

    Description:
        This value type might be used to reference binary
        information, for values that are large, or otherwise undesirable
        to include directly in the iCalendar object.

        Property values with this value type MUST follow the generic URI
        syntax defined in [RFC3986].

        When a property parameter value is a URI value type, the URI MUST
        be specified as a quoted-string value.

    Examples:
        The following is a URI for a network file:

        .. code-block:: ics

            http://example.com/my-report.txt

        .. code-block:: pycon

            >>> from icalendar.prop import vUri
            >>> uri = vUri.from_ical('http://example.com/my-report.txt')
            >>> uri
            vUri('http://example.com/my-report.txt')
            >>> uri.uri
            'http://example.com/my-report.txt'
    """

    default_value: ClassVar[str] = "URI"
    params: Parameters
    __slots__ = ("params",)

    def __new__(
        cls,
        value: str,
        encoding: str = DEFAULT_ENCODING,
        /,
        params: dict[str, Any] | None = None,
    ) -> Self:
        value = to_unicode(value, encoding=encoding)
        self = super().__new__(cls, value)
        self.params = Parameters(params)
        return self

    def to_ical(self) -> bytes:
        return self.encode(DEFAULT_ENCODING)

    @classmethod
    def from_ical(cls, ical: str | bytes) -> Self:
        try:
            return cls(ical)
        except Exception as e:
            raise ValueError(f"Expected , got: {ical}") from e

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vUri."""
        return [cls("http://example.com/my-report.txt")]

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        return [name, self.params.to_jcal(), self.VALUE.lower(), str(self)]

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        return cls(
            jcal_property[3],
            Parameters.from_jcal_property(jcal_property),
        )

    @property
    def ical_value(self) -> str:
        """The URI."""
        return self.uri

    @property
    def uri(self) -> str:
        """The URI."""
        return str(self)

    def __repr__(self) -> str:
        """repr(self)"""
        return f"{self.__class__.__name__}({self.uri!r})"

    from icalendar.param import FMTTYPE, GAP, LABEL, LANGUAGE, LINKREL, RELTYPE, VALUE


__all__ = ["vUri"]



"""parsing and generation of content lines"""

import re

from icalendar.parser.parameter import Parameters
from icalendar.parser.property import unescape_backslash, unescape_list_or_string
from icalendar.parser.string import (
    _escape_string,
    _unescape_string,
    foldline,
    validate_token,
)
from icalendar.parser_tools import DEFAULT_ENCODING, ICAL_TYPE, to_unicode

UFOLD = re.compile("(\r?\n)+[ \t]")
NEWLINE = re.compile(r"\r?\n")

OWS = " \t"
OWS_AROUND_DELIMITERS_RE = re.compile(r"[ \t]*([;=])[ \t]*")


def _strip_ows_around_delimiters(st: str, delimiters: str = ";=") -> str:
    """Strip optional whitespace around delimiters outside of quoted sections,
    respecting backslash escapes so that escaped delimiters are not treated as
    separators.

    This is a lenient parsing helper (used when strict=False) to support
    iCalendar content lines that contain extra whitespace around tokens.
    """
    if not st:
        return st

    # Fast path for the common case in non-strict mode:
    # no whitespace in the parameter section means there is nothing to normalize.
    if " " not in st and "\t" not in st:
        return st

    # Fast regex-based path for simple parameter sections without quoting/escaping.
    if delimiters == ";=" and '"' not in st and "\\" not in st:
        return OWS_AROUND_DELIMITERS_RE.sub(r"\1", st).strip()

    out: list[str] = []
    pending_ws: list[str] = []
    in_quotes = False
    escaped = False
    # True only if the last appended char was a raw delimiter.
    last_was_delimiter = False

    def flush_pending() -> None:
        nonlocal pending_ws
        if not pending_ws:
            return
        if not last_was_delimiter:
            out.extend(pending_ws)
        pending_ws.clear()

    for ch in st:
        # Handle escaped character (the backslash set escaped in previous iteration)
        if escaped:
            flush_pending()
            out.append(ch)
            escaped = False
            last_was_delimiter = False
            continue

        # Handle backslash to escape next character
        if ch == "\\" and not in_quotes:
            flush_pending()
            out.append(ch)
            escaped = True
            last_was_delimiter = False
            continue

        # Handle quote toggling
        if ch == '"' and not escaped:
            in_quotes = not in_quotes
            flush_pending()
            out.append(ch)
            last_was_delimiter = False
            continue

        # Whitespace outside quotes is buffered
        if not in_quotes and not escaped and ch in OWS:
            pending_ws.append(ch)
            continue

        # Raw delimiter (unescaped and outside quotes)
        if not in_quotes and not escaped and ch in delimiters:
            pending_ws.clear()
            while out and out[-1] in OWS:
                out.pop()
            out.append(ch)
            last_was_delimiter = True
            continue

        # Regular character
        flush_pending()
        out.append(ch)
        last_was_delimiter = False

    if pending_ws and not last_was_delimiter:
        out.extend(pending_ws)

    return "".join(out).strip()


class Contentline(str):
    """A content line is basically a string that can be folded and parsed into
    parts.
    """

    __slots__ = ("strict",)

    def __new__(cls, value, strict=False, encoding=DEFAULT_ENCODING):
        value = to_unicode(value, encoding=encoding)
        assert "\n" not in value, (
            "Content line can not contain unescaped new line characters."
        )
        self = super().__new__(cls, value)
        self.strict = strict
        return self

    @classmethod
    def from_parts(
        cls,
        name: ICAL_TYPE,
        params: Parameters,
        values,
        sorted: bool = True,  # noqa: A002
    ):
        """Turn a parts into a content line."""
        assert isinstance(params, Parameters)
        if hasattr(values, "to_ical"):
            values = values.to_ical()
        else:
            from icalendar.prop import vText

            values = vText(values).to_ical()
        # elif isinstance(values, basestring):
        #    values = escape_char(values)

        # TODO: after unicode only, remove this
        # Convert back to unicode, after to_ical encoded it.
        name = to_unicode(name)
        values = to_unicode(values)
        if params:
            params = to_unicode(params.to_ical(sorted=sorted))
            if params:
                # some parameter values can be skipped during serialization
                return cls(f"{name};{params}:{values}")
        return cls(f"{name}:{values}")

    def parts(self) -> tuple[str, Parameters, str]:
        """Split the content line into ``name``, ``parameters``, and ``values`` parts.

        Properly handles escaping with backslashes and double-quote sections
        to avoid corrupting URL-encoded characters in values.

        Example with parameter:

        .. code-block:: ics

            DESCRIPTION;ALTREP="cid:part1.0001@example.org":The Fall'98 Wild

        Example without parameters:

        .. code-block:: ics

            DESCRIPTION:The Fall'98 Wild
        """
        try:
            name_split: int | None = None
            value_split: int | None = None
            in_quotes: bool = False
            escaped: bool = False

            for i, ch in enumerate(self):
                if ch == '"' and not escaped:
                    in_quotes = not in_quotes
                elif ch == "\\" and not in_quotes:
                    escaped = True
                    continue
                elif not in_quotes and not escaped:
                    # Find first delimiter for name
                    if ch in ":;" and name_split is None:
                        name_split = i
                    # Find value delimiter (first colon)
                    if ch == ":" and value_split is None:
                        value_split = i

                escaped = False

            # Validate parsing results
            if not value_split:
                # No colon found - value is empty, use end of string
                value_split = len(self)

            # Extract name - if no delimiter,
            #   take whole string for validate_token to reject
            name = self[:name_split] if name_split else self
            if not self.strict:
                name = re.sub(r"[ \t]+", "", name.strip())
            validate_token(name)

            if not name_split or name_split + 1 == value_split:
                # No delimiter or empty parameter section
                raise ValueError("Invalid content line")  # noqa: TRY301
            # Parse parameters - they still need to be escaped/unescaped
            # for proper handling of commas, semicolons, etc. in parameter values
            raw_param_str = self[name_split + 1 : value_split]
            if not self.strict:
                raw_param_str = _strip_ows_around_delimiters(raw_param_str)
            param_str = _escape_string(raw_param_str)
            params = Parameters.from_ical(param_str, strict=self.strict)
            params = Parameters(
                (_unescape_string(key), unescape_list_or_string(value))
                for key, value in iter(params.items())
            )
            # Unescape backslash sequences in values but preserve URL encoding
            values = unescape_backslash(self[value_split + 1 :])
        except ValueError as exc:
            raise ValueError(
                f"Content line could not be parsed into parts: '{self}': {exc}"
            ) from exc
        return (name, params, values)

    @classmethod
    def from_ical(cls, ical, strict=False):
        """Unfold the content lines in an iCalendar into long content lines."""
        ical = to_unicode(ical)
        # a fold is carriage return followed by either a space or a tab
        return cls(UFOLD.sub("", ical), strict=strict)

    def to_ical(self):
        """Long content lines are folded so they are less than 75 characters
        wide.
        """
        return foldline(self).encode(DEFAULT_ENCODING)


class Contentlines(list[Contentline]):
    """I assume that iCalendar files generally are a few kilobytes in size.
    Then this should be efficient. for Huge files, an iterator should probably
    be used instead.
    """

    def to_ical(self):
        """Simply join self."""
        return b"\r\n".join(line.to_ical() for line in self if line) + b"\r\n"

    @classmethod
    def from_ical(cls, st):
        """Parses a string into content lines."""
        st = to_unicode(st)
        try:
            # a fold is carriage return followed by either a space or a tab
            unfolded = UFOLD.sub("", st)
            lines = cls(Contentline(line) for line in NEWLINE.split(unfolded) if line)
            lines.append("")  # '\r\n' at the end of every content line
        except Exception as e:
            raise ValueError("Expected StringType with content lines") from e
        return lines


__all__ = ["Contentline", "Contentlines"]



"""INT values from :rfc:`5545`."""

from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.parser_tools import ICAL_TYPE


class vInt(int):
    """Integer

    Value Name:
        INTEGER

    Purpose:
        This value type is used to identify properties that contain a
        signed integer value.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            integer    = (["+"] / "-") 1*DIGIT

    Description:
        If the property permits, multiple "integer" values are
        specified by a COMMA-separated list of values.  The valid range
        for "integer" is -2147483648 to 2147483647.  If the sign is not
        specified, then the value is assumed to be positive.

    The ``__new__`` method creates a vInt instance:

    Parameters:
        value: Integer value to encode. Can be positive or negative within
            the range -2147483648 to 2147483647.
        params: Optional parameter dictionary for the property.

    Returns:
        vInt instance

    Examples:

        .. code-block:: text

            1234567890
            -1234567890
            +1234567890
            432109876

        .. code-block:: pycon

            >>> from icalendar.prop import vInt
            >>> integer = vInt.from_ical('1234567890')
            >>> integer
            1234567890
            >>> integer = vInt.from_ical('-1234567890')
            >>> integer
            -1234567890
            >>> integer = vInt.from_ical('+1234567890')
            >>> integer
            1234567890
            >>> integer = vInt.from_ical('432109876')
            >>> integer
            432109876

        Create a PRIORITY property (1 = highest priority):

        .. code-block:: pycon

            >>> priority = vInt(1)
            >>> priority
            1
            >>> priority.to_ical()
            b'1'

        Create SEQUENCE property (for versioning):

        .. code-block:: pycon

            >>> sequence = vInt(3)
            >>> sequence.to_ical()
            b'3'
    """

    default_value: ClassVar[str] = "INTEGER"
    params: Parameters

    def __new__(cls, *args, params: dict[str, Any] | None = None, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self.params = Parameters(params)
        return self

    def to_ical(self) -> bytes:
        return str(self).encode("utf-8")

    @property
    def ical_value(self) -> int:
        """INTEGER property type according to :rfc:`5545#section-3.3.8`"""
        return int(self)

    @classmethod
    def from_ical(cls, ical: ICAL_TYPE):
        try:
            return cls(ical)
        except Exception as e:
            raise ValueError(f"Expected int, got: {ical}") from e

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vInt."""
        return [vInt(1000), vInt(-42)]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        return [name, self.params.to_jcal(), self.VALUE.lower(), int(self)]

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        JCalParsingError.validate_value_type(jcal_property[3], int, cls, 3)
        return cls(
            jcal_property[3],
            params=Parameters.from_jcal_property(jcal_property),
        )

    @classmethod
    def parse_jcal_value(cls, value: Any) -> int:
        """Parse a jCal value for vInt.

        Raises:
            ~error.JCalParsingError: If the value is not an int.
        """
        JCalParsingError.validate_value_type(value, int, cls)
        return cls(value)


__all__ = ["vInt"]



"""ADR property of :rfc:`6350`."""

from typing import Any, ClassVar, NamedTuple

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.parser_tools import DEFAULT_ENCODING, to_unicode


class AdrFields(NamedTuple):
    """Named fields for vCard ADR (Address) property per :rfc:`6350#section-6.3.1`.

    Provides named access to the seven address components.
    """

    po_box: str
    """Post office box."""
    extended: str
    """Extended address (e.g., apartment or suite number)."""
    street: str
    """Street address."""
    locality: str
    """Locality (e.g., city)."""
    region: str
    """Region (e.g., state or province)."""
    postal_code: str
    """Postal code."""
    country: str
    """Country name (full name)."""


class vAdr:
    """vCard ADR (Address) structured property per :rfc:`6350#section-6.3.1`.

    The ADR property represents a delivery address as a single text value.
    The structured type value consists of a sequence of seven address components.
    The component values must be specified in their corresponding position.

    -   post office box
    -   extended address (e.g., apartment or suite number)
    -   street address
    -   locality (e.g., city)
    -   region (e.g., state or province)
    -   postal code
    -   country name (full name)

    When a component value is missing, the associated component separator
    MUST still be specified.

    Semicolons are field separators and are NOT escaped.
    Commas and backslashes within field values ARE escaped per :rfc:`6350`.

    Examples:
        .. code-block:: pycon

            >>> from icalendar.prop import vAdr
            >>> adr = vAdr(("", "", "123 Main St", "Springfield", "IL", "62701", "USA"))
            >>> adr.to_ical()
            b';;123 Main St;Springfield;IL;62701;USA'
            >>> vAdr.from_ical(";;123 Main St;Springfield;IL;62701;USA")
            AdrFields(po_box='', extended='', street='123 Main St', locality='Springfield', region='IL', postal_code='62701', country='USA')
    """

    default_value: ClassVar[str] = "TEXT"
    params: Parameters
    fields: AdrFields

    def __init__(
        self,
        fields: tuple[str, ...] | list[str] | str | AdrFields,
        /,
        params: dict[str, Any] | None = None,
    ):
        """Initialize ADR with seven fields or parse from vCard format string.

        Parameters:
            fields: Either an AdrFields, tuple, or list of seven strings, one per field,
                    or a vCard format string with semicolon-separated fields
            params: Optional property parameters
        """
        if isinstance(fields, str):
            fields = self.from_ical(fields)
        if isinstance(fields, AdrFields):
            self.fields = fields
        else:
            if len(fields) != 7:
                raise ValueError(f"ADR must have exactly 7 fields, got {len(fields)}")
            self.fields = AdrFields(*(str(f) for f in fields))
        self.params = Parameters(params)

    def to_ical(self) -> bytes:
        """Generate vCard format with semicolon-separated fields."""
        # Each field is vText (handles comma/backslash escaping)
        # but we join with unescaped semicolons (field separators)
        from icalendar.prop.text import vText

        parts = [vText(f).to_ical().decode(DEFAULT_ENCODING) for f in self.fields]
        return ";".join(parts).encode(DEFAULT_ENCODING)

    @staticmethod
    def from_ical(ical: str | bytes) -> AdrFields:
        """Parse vCard ADR format into an AdrFields named tuple.

        Parameters:
            ical: vCard format string with semicolon-separated fields

        Returns:
            AdrFields named tuple with seven field values.
        """
        from icalendar.parser import split_on_unescaped_semicolon

        ical = to_unicode(ical)
        fields = split_on_unescaped_semicolon(ical)
        if len(fields) != 7:
            raise ValueError(
                f"ADR must have exactly 7 fields, got {len(fields)}: {ical}"
            )
        return AdrFields(*fields)

    def __eq__(self, other: object) -> bool:
        """self == other"""
        return isinstance(other, vAdr) and self.fields == other.fields

    def __repr__(self) -> str:
        """String representation."""
        return f"{self.__class__.__name__}({self.fields}, params={self.params})"

    @property
    def ical_value(self) -> AdrFields:
        """The address fields as a named tuple."""
        return self.fields

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        result = [name, self.params.to_jcal(), self.VALUE.lower()]
        result.extend(self.fields)
        return result

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        if len(jcal_property) != 10:  # name, params, value_type, 7 fields
            raise JCalParsingError(
                f"ADR must have 10 elements (name, params, value_type, 7 fields), "
                f"got {len(jcal_property)}"
            )
        for i, field in enumerate(jcal_property[3:], start=3):
            JCalParsingError.validate_value_type(field, str, cls, i)
        return cls(
            tuple(jcal_property[3:]),
            Parameters.from_jcal_property(jcal_property),
        )

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vAdr."""
        return [cls(("", "", "123 Main St", "Springfield", "IL", "62701", "USA"))]


__all__ = ["AdrFields", "vAdr"]



"""This module contains the parser/generators (or coders/encoders if you
prefer) for the classes/datatypes that are used in iCalendar:

###########################################################################

# This module defines these property value data types and property parameters

4.2 Defined property parameters are:

.. code-block:: ics

     ALTREP, CN, CUTYPE, DELEGATED-FROM, DELEGATED-TO, DIR, ENCODING, FMTTYPE,
     FBTYPE, LANGUAGE, MEMBER, PARTSTAT, RANGE, RELATED, RELTYPE, ROLE, RSVP,
     SENT-BY, TZID, VALUE

4.3 Defined value data types are:

.. code-block:: ics

    BINARY, BOOLEAN, CAL-ADDRESS, DATE, DATE-TIME, DURATION, FLOAT, INTEGER,
    PERIOD, RECUR, TEXT, TIME, URI, UTC-OFFSET

###########################################################################

iCalendar properties have values. The values are strongly typed. This module
defines these types, calling val.to_ical() on them will render them as defined
in rfc5545.

If you pass any of these classes a Python primitive, you will have an object
that can render itself as iCalendar formatted date.

Property Value Data Types start with a 'v'. they all have an to_ical() and
from_ical() method. The to_ical() method generates a text string in the
iCalendar format. The from_ical() method can parse this format and return a
primitive Python datatype. So it should always be true that:

.. code-block:: python

    x == vDataType.from_ical(VDataType(x).to_ical())

These types are mainly used for parsing and file generation. But you can set
them directly.
"""

from typing import TypeAlias

from .adr import AdrFields, vAdr
from .binary import vBinary
from .boolean import vBoolean
from .broken import vBroken
from .cal_address import vCalAddress
from .categories import vCategory
from .dt import (
    DT_TYPE,
    TimeBase,
    vDate,
    vDatetime,
    vDDDLists,
    vDDDTypes,
    vDuration,
    vPeriod,
    vTime,
    vUTCOffset,
)
from .factory import TypesFactory
from .float import vFloat
from .geo import vGeo
from .inline import vInline
from .integer import vInt
from .n import NFields, vN
from .org import vOrg
from .recur import vFrequency, vMonth, vRecur, vSkip, vWeekday
from .text import vText
from .uid import vUid
from .unknown import vUnknown
from .uri import vUri
from .xml_reference import vXmlReference

VPROPERTY: TypeAlias = (
    vAdr
    | vBoolean
    | vBroken
    | vCalAddress
    | vCategory
    | vDDDLists
    | vDDDTypes
    | vDate
    | vDatetime
    | vDuration
    | vFloat
    | vFrequency
    | vInt
    | vMonth
    | vN
    | vOrg
    | vPeriod
    | vRecur
    | vSkip
    | vText
    | vTime
    | vUTCOffset
    | vUri
    | vWeekday
    | vInline
    | vBinary
    | vGeo
    | vUnknown
    | vXmlReference
    | vUid
)

__all__ = [
    "DT_TYPE",
    "VPROPERTY",
    "AdrFields",
    "NFields",
    "TimeBase",
    "TypesFactory",
    "vAdr",
    "vBinary",
    "vBoolean",
    "vBroken",
    "vCalAddress",
    "vCategory",
    "vDDDLists",
    "vDDDTypes",
    "vDate",
    "vDatetime",
    "vDuration",
    "vFloat",
    "vFrequency",
    "vGeo",
    "vInline",
    "vInt",
    "vMonth",
    "vN",
    "vOrg",
    "vPeriod",
    "vRecur",
    "vSkip",
    "vText",
    "vTime",
    "vUTCOffset",
    "vUid",
    "vUnknown",
    "vUri",
    "vWeekday",
    "vXmlReference",
]



"""Special parsing for calendar components."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from icalendar.parser.content_line import Contentline

from .component import ComponentIcalParser

if TYPE_CHECKING:
    from icalendar.cal.component import Component


class LazyCalendarIcalParser(ComponentIcalParser):
    """A parser for calendar components.

    Instead of parsing the components, LazyComponents are created.
    Parsing can happen on demand.

    A calendar may grow over time, with its subcomponents greatly increasing
    in number, and requiring more memory and time to fully parse. This
    optimization lazily parses calendar files without consuming more memory
    than necessary, reducing the initial time it takes to access meta data.
    """

    parse_instantly: ClassVar[tuple[str, ...]] = ("VCALENDAR",)
    """Parse these components immediately, instead of lazily.

    All other components are parsed lazily.
    """

    def handle_begin_component(self, vals):
        """Begin a new component.

        This may be the first component.
        """
        c_name = vals.upper()
        if (
            c_name in self.parse_instantly
            or not self.component
            or not self.component.is_lazy()
        ):
            # these components are parsed immediately
            super().handle_begin_component(vals)
        else:
            self.handle_lazy_begin_component(c_name)

    def handle_lazy_begin_component(self, component_name: str) -> None:
        """Begin a new component, but do not parse it yet.

        Parameters:
            component_name:
                The upper case name of the component, for example, ``"VEVENT"``.
        """
        content_lines = [Contentline(f"BEGIN:{component_name}")]
        for line in self._content_lines_iterator:
            content_lines.append(line)
            if (
                line[:4].upper() == "END:"
                and line[4:].strip().upper() == component_name
            ):
                break
        if self.component is None:
            raise ValueError(
                f"BEGIN:{component_name} encountered outside of a parent component."
            )
        self.component.add_component(
            LazySubcomponent(
                component_name,
                self.get_subcomponent_parser(content_lines),
            )
        )

    def get_subcomponent_parser(
        self, content_lines: list[Contentline]
    ) -> ComponentIcalParser:
        """Get the parser for a subcomponent.

        Parameters:
            content_lines: The content lines of the subcomponent.
        """
        return ComponentIcalParser(
            content_lines, self._component_factory, self._types_factory
        )

    def prepare_components(self):
        """Prepare the lazily parsed components."""


class LazySubcomponent:
    """A subcomponent that is evaluated lazily.

    This class holds the raw data of the subcomponent ready for parsing.
    """

    def __init__(self, name: str, parser: ComponentIcalParser):
        """Initialize the lazy subcomponent with the raw data."""
        self._name = name
        self._parser = parser
        self._component: Component | None = None

    @property
    def name(self) -> str:
        """The name of the subcomponent.

        The name is uppercased, per :rfc:`5545#section-2.1`.
        """
        return self._name

    def is_parsed(self) -> bool:
        """Return whether the subcomponent is already parsed."""
        return self._component is not None

    def parse(self) -> Component:
        """Parse the raw data and return the component."""
        if self._component is None:
            components = self._parser.parse()
            if len(components) != 1:
                raise ValueError(
                    f"Expected exactly one component in the subcomponent, "
                    f"but got {len(components)}."
                )
            self._component = components[0]
            self._parser = None  # free memory
        return self._component

    def is_lazy(self) -> bool:
        """Return whether the subcomponents were accessed and parsed lazily.

        Call :meth:`parse` to get the fully parsed component.
        """
        return True

    def __repr__(self) -> str:
        return f"LazySubcomponent(name={self._name}, parsed={self.is_parsed()})"

    def walk(self, name: str) -> list[Component]:
        """Walk through this component.

        This only parses the component if necessary.

        Parameters:
            name: The name to match for all components in the calendar,
                then walk through and parse the resulting matches.
        """
        if not isinstance(name, str):
            raise TypeError("name must be a string.")
        if name == self.name or (
            self._parser and self._parser.contains_component(name)
        ):
            return self.parse().walk(name)
        return []

    def with_uid(self, uid: str) -> list[Component]:
        """Return the components containing the given ``uid``.

        This only parses the component if necessary.

        Parameters:
            uid: The UID of the components.
        """
        if self._parser and not self._parser.contains_uid(uid):
            return []
        return self.parse().with_uid(uid)


__all__ = ["LazyCalendarIcalParser", "LazySubcomponent"]



"""SKIP value type of RECUR from :rfc:`7529`."""

from typing import Any

from icalendar.compatibility import Self
from icalendar.enums import Enum
from icalendar.error import JCalParsingError
from icalendar.prop.text import vText


class vSkip(vText, Enum):
    """Skip values for RRULE.

    These are defined in :rfc:`7529`.

    OMIT  is the default value.

    Examples:

    .. code-block:: pycon

        >>> from icalendar import vSkip
        >>> vSkip.OMIT
        vSkip('OMIT')
        >>> vSkip.FORWARD
        vSkip('FORWARD')
        >>> vSkip.BACKWARD
        vSkip('BACKWARD')
    """

    OMIT = "OMIT"
    FORWARD = "FORWARD"
    BACKWARD = "BACKWARD"

    __reduce_ex__ = Enum.__reduce_ex__

    def __repr__(self):
        return f"{self.__class__.__name__}({self._name_!r})"

    @classmethod
    def parse_jcal_value(cls, value: Any) -> Self:
        """Parse a jCal value for vSkip.

        Raises:
            ~error.JCalParsingError: If the value is not a valid skip value.
        """
        JCalParsingError.validate_value_type(value, str, cls)
        try:
            return cls[value.upper()]
        except KeyError as e:
            raise JCalParsingError(
                "The value must be a valid skip value.", cls, value=value
            ) from e


__all__ = ["vSkip"]



"""RECUR property type from :rfc:`5545`."""

from typing import Any, ClassVar

from icalendar.caselessdict import CaselessDict
from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.parser_tools import DEFAULT_ENCODING, SEQUENCE_TYPES
from icalendar.prop.dt import vDDDTypes
from icalendar.prop.integer import vInt
from icalendar.prop.recur.frequency import vFrequency
from icalendar.prop.recur.month import vMonth
from icalendar.prop.recur.skip import vSkip
from icalendar.prop.recur.weekday import vWeekday
from icalendar.prop.text import vText


class vRecur(CaselessDict):
    """Recurrence definition.

    Property Name:
        RRULE

    Purpose:
        This property defines a rule or repeating pattern for recurring events, to-dos,
        journal entries, or time zone definitions.

    Value Type:
        RECUR

    Property Parameters:
        IANA and non-standard property parameters can be specified on this property.

    Conformance:
        This property can be specified in recurring "VEVENT", "VTODO", and "VJOURNAL"
        calendar components as well as in the "STANDARD" and "DAYLIGHT" sub-components
        of the "VTIMEZONE" calendar component, but it SHOULD NOT be specified more than once.
        The recurrence set generated with multiple "RRULE" properties is undefined.

    Description:
        The recurrence rule, if specified, is used in computing the recurrence set.
        The recurrence set is the complete set of recurrence instances for a calendar component.
        The recurrence set is generated by considering the initial "DTSTART" property along
        with the "RRULE", "RDATE", and "EXDATE" properties contained within the
        recurring component. The "DTSTART" property defines the first instance in the
        recurrence set. The "DTSTART" property value SHOULD be synchronized with the
        recurrence rule, if specified. The recurrence set generated with a "DTSTART" property
        value not synchronized with the recurrence rule is undefined.
        The final recurrence set is generated by gathering all of the start DATE-TIME
        values generated by any of the specified "RRULE" and "RDATE" properties, and then
        excluding any start DATE-TIME values specified by "EXDATE" properties.
        This implies that start DATE- TIME values specified by "EXDATE" properties take
        precedence over those specified by inclusion properties (i.e., "RDATE" and "RRULE").
        Where duplicate instances are generated by the "RRULE" and "RDATE" properties,
        only one recurrence is considered. Duplicate instances are ignored.

        The "DTSTART" property specified within the iCalendar object defines the first
        instance of the recurrence. In most cases, a "DTSTART" property of DATE-TIME value
        type used with a recurrence rule, should be specified as a date with local time
        and time zone reference to make sure all the recurrence instances start at the
        same local time regardless of time zone changes.

        If the duration of the recurring component is specified with the "DTEND" or
        "DUE" property, then the same exact duration will apply to all the members of the
        generated recurrence set. Else, if the duration of the recurring component is
        specified with the "DURATION" property, then the same nominal duration will apply
        to all the members of the generated recurrence set and the exact duration of each
        recurrence instance will depend on its specific start time. For example, recurrence
        instances of a nominal duration of one day will have an exact duration of more or less
        than 24 hours on a day where a time zone shift occurs. The duration of a specific
        recurrence may be modified in an exception component or simply by using an
        "RDATE" property of PERIOD value type.

    Examples:
        The following RRULE specifies daily events for 10 occurrences.

        .. code-block:: ics

            RRULE:FREQ=DAILY;COUNT=10

        Below, we parse the RRULE ical string.

        .. code-block:: pycon

            >>> from icalendar.prop import vRecur
            >>> rrule = vRecur.from_ical('FREQ=DAILY;COUNT=10')
            >>> rrule
            vRecur({'FREQ': ['DAILY'], 'COUNT': [10]})

        You can choose to add an rrule to an :class:`icalendar.cal.Event` or
        :class:`icalendar.cal.Todo`.

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> event = Event()
            >>> event.add('RRULE', 'FREQ=DAILY;COUNT=10')
            >>> event.rrules
            [vRecur({'FREQ': ['DAILY'], 'COUNT': [10]})]
    """

    default_value: ClassVar[str] = "RECUR"
    params: Parameters

    frequencies = [
        "SECONDLY",
        "MINUTELY",
        "HOURLY",
        "DAILY",
        "WEEKLY",
        "MONTHLY",
        "YEARLY",
    ]

    # Mac iCal ignores RRULEs where FREQ is not the first rule part.
    # Sorts parts according to the order listed in RFC 5545, section 3.3.10.
    canonical_order = (
        "RSCALE",
        "FREQ",
        "UNTIL",
        "COUNT",
        "INTERVAL",
        "BYSECOND",
        "BYMINUTE",
        "BYHOUR",
        "BYDAY",
        "BYWEEKDAY",
        "BYMONTHDAY",
        "BYYEARDAY",
        "BYWEEKNO",
        "BYMONTH",
        "BYSETPOS",
        "WKST",
        "SKIP",
    )

    types = CaselessDict(
        {
            "COUNT": vInt,
            "INTERVAL": vInt,
            "BYSECOND": vInt,
            "BYMINUTE": vInt,
            "BYHOUR": vInt,
            "BYWEEKNO": vInt,
            "BYMONTHDAY": vInt,
            "BYYEARDAY": vInt,
            "BYMONTH": vMonth,
            "UNTIL": vDDDTypes,
            "BYSETPOS": vInt,
            "WKST": vWeekday,
            "BYDAY": vWeekday,
            "FREQ": vFrequency,
            "BYWEEKDAY": vWeekday,
            "SKIP": vSkip,  # RFC 7529
            "RSCALE": vText,  # RFC 7529
        }
    )

    # for reproducible serialization:
    # RULE: if and only if it can be a list it will be a list
    # look up in RFC
    jcal_not_a_list = {"FREQ", "UNTIL", "COUNT", "INTERVAL", "WKST", "SKIP", "RSCALE"}

    def __init__(self, *args, params: dict[str, Any] | None = None, **kwargs):
        if args and isinstance(args[0], str):
            # we have a string as an argument.
            args = (self.from_ical(args[0]),) + args[1:]
        for k, v in kwargs.items():
            if not isinstance(v, SEQUENCE_TYPES):
                kwargs[k] = [v]
        super().__init__(*args, **kwargs)
        self.params = Parameters(params)

    def to_ical(self):
        result = []
        for key, vals in self.sorted_items():
            typ = self.types.get(key, vText)
            if not isinstance(vals, SEQUENCE_TYPES):
                vals = [vals]
            param_vals = b",".join(typ(val).to_ical() for val in vals)

            # CaselessDict keys are always unicode
            param_key = key.encode(DEFAULT_ENCODING)
            result.append(param_key + b"=" + param_vals)

        return b";".join(result)

    @classmethod
    def parse_type(cls, key, values):
        # integers
        parser = cls.types.get(key, vText)
        return [parser.from_ical(v) for v in values.split(",")]

    @classmethod
    def from_ical(cls, ical: str):
        if isinstance(ical, cls):
            return ical
        try:
            recur = cls()
            for pairs in ical.split(";"):
                try:
                    key, vals = pairs.split("=")
                except ValueError:
                    # E.g. incorrect trailing semicolon, like (issue #157):
                    # FREQ=YEARLY;BYMONTH=11;BYDAY=1SU;
                    continue
                recur[key] = cls.parse_type(key, vals)
            return cls(recur)
        except ValueError:
            raise
        except Exception as e:
            raise ValueError(f"Error in recurrence rule: {ical}") from e

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vRecur."""
        return [cls.from_ical("FREQ=DAILY;COUNT=10")]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        recur = {}
        for k, v in self.items():
            key = k.lower()
            if key.upper() in self.jcal_not_a_list:
                value = v[0] if isinstance(v, list) and len(v) == 1 else v
            elif not isinstance(v, list):
                value = [v]
            else:
                value = v
            recur[key] = value
        if "until" in recur:
            until = recur["until"]
            until_jcal = vDDDTypes(until).to_jcal("until")
            recur["until"] = until_jcal[-1]
        return [name, self.params.to_jcal(), self.VALUE.lower(), recur]

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        params = Parameters.from_jcal_property(jcal_property)
        if not isinstance(jcal_property[3], dict) or not all(
            isinstance(k, str) for k in jcal_property[3]
        ):
            raise JCalParsingError(
                "The recurrence rule must be a mapping with string keys.",
                cls,
                3,
                value=jcal_property[3],
            )
        recur = {}
        for key, value in jcal_property[3].items():
            value_type = cls.types.get(key, vText)
            with JCalParsingError.reraise_with_path_added(3, key):
                if isinstance(value, list):
                    recur[key.lower()] = values = []
                    for i, v in enumerate(value):
                        with JCalParsingError.reraise_with_path_added(i):
                            values.append(value_type.parse_jcal_value(v))
                else:
                    recur[key] = value_type.parse_jcal_value(value)
        until = recur.get("until")
        if until is not None and not isinstance(until, list):
            recur["until"] = [until]
        return cls(recur, params=params)

    def __eq__(self, other: object) -> bool:
        """self == other"""
        if not isinstance(other, vRecur):
            return super().__eq__(other)
        if self.keys() != other.keys():
            return False
        for key in self.keys():
            v1 = self[key]
            v2 = other[key]
            if not isinstance(v1, SEQUENCE_TYPES):
                v1 = [v1]
            if not isinstance(v2, SEQUENCE_TYPES):
                v2 = [v2]
            if v1 != v2:
                return False
        return True

    __hash__ = None


__all__ = ["vRecur"]



"""TEXT values from :rfc:`5545`."""

from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters, _escape_char
from icalendar.parser_tools import DEFAULT_ENCODING, ICAL_TYPE, to_unicode


class vText(str):
    r"""vText is a data type that contains human-readable text values.

    The vText property uses the :rfc:`5545#section-3.3.11` TEXT value type
    in various icalendar properties to show free-form text that others can read.
    This class can be created from Python strings, and can be used to add text
    descriptions to calendar events.

    To create a TEXT object, pass in the string you want when creating the
    object.

    To add a line break, use ``\n`` or ``\N``.

    Use the LANGUAGE property parameter to set the language of the text.

    When the TEXT object is serialized to an icalendar stream, certain
    characters are escaped or changed.
    These characters include the COMMA, SEMICOLON, BACKSLASH, and line breaks.

    Examples:

        vText property as a TEXT value type.

        .. code-block:: text

            Project XYZ Final Review\nConference Room - 3B\nCome Prepared.

        Create a vText property, and display it in a readable format.

        .. code-block:: pycon

            >>> from icalendar.prop import vText
            >>> desc = 'Project XYZ Final Review\nConference Room - 3B\nCome Prepared.'
            >>> text = vText(desc)
            >>> text
            vText(b'Project XYZ Final Review\\nConference Room - 3B\\nCome Prepared.')
            >>> print(text.ical_value)
            Project XYZ Final Review
            Conference Room - 3B
            Come Prepared.

        Add a SUMMARY to an event, then display its value as a vText property then in a readable format:

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> event = Event()
            >>> event.add('SUMMARY', desc)
            >>> event['SUMMARY']
            vText(b'Project XYZ Final Review\\nConference Room - 3B\\nCome Prepared.')
            >>> print(event.to_ical().decode())
            BEGIN:VEVENT
            SUMMARY:Project XYZ Final Review\nConference Room - 3B\nCome Prepared.
            END:VEVENT

    """

    default_value: ClassVar[str] = "TEXT"
    params: Parameters
    __slots__ = ("encoding", "params")

    def __new__(
        cls,
        value: ICAL_TYPE,
        encoding: str = DEFAULT_ENCODING,
        /,
        params: dict[str, Any] | None = None,
    ) -> Self:
        value = to_unicode(value, encoding=encoding)
        self = super().__new__(cls, value)
        self.encoding = encoding
        self.params = Parameters(params)
        return self

    def __repr__(self) -> str:
        return f"vText({self.to_ical()!r})"

    def to_ical(self) -> bytes:
        return _escape_char(self).encode(self.encoding)

    @classmethod
    def from_ical(cls, ical: ICAL_TYPE) -> Self:
        return cls(ical)

    @property
    def ical_value(self) -> str:
        """The string value of the text."""
        return str(self)

    from icalendar.param import ALTREP, GAP, LANGUAGE, RELTYPE, VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        if name == "request-status":  # TODO: maybe add a vRequestStatus class?
            return [name, {}, "text", self.split(";", 2)]
        return [name, self.params.to_jcal(), self.VALUE.lower(), str(self)]

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vText."""
        return [cls("Hello World!")]

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        name = jcal_property[0]
        if name == "categories":
            from icalendar.prop import vCategory

            return vCategory.from_jcal(jcal_property)
        string = jcal_property[3]  # TODO: accept list or string but join with ;
        if name == "request-status":  # TODO: maybe add a vRequestStatus class?
            JCalParsingError.validate_list_type(jcal_property[3], str, cls, 3)
            string = ";".join(jcal_property[3])
        JCalParsingError.validate_value_type(string, str, cls, 3)
        return cls(
            string,
            params=Parameters.from_jcal_property(jcal_property),
        )

    @classmethod
    def parse_jcal_value(cls, jcal_value: Any) -> Self:
        """Parse a jCal value into a vText."""
        JCalParsingError.validate_value_type(jcal_value, (str, int, float), cls)
        return cls(str(jcal_value))


__all__ = ["vText"]



"""DATE-TIME property type from :rfc:`5545`."""

from datetime import datetime
from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.timezone import tzp
from icalendar.timezone.tzid import is_utc

from .base import TimeBase


class vDatetime(TimeBase):
    """Date-Time

    Value Name:
        DATE-TIME

    Purpose:
        This value type is used to identify values that specify a
        precise calendar date and time of day. The format is based on
        the ISO.8601.2004 complete representation.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            date-time  = date "T" time

            date       = date-value
            date-value         = date-fullyear date-month date-mday
            date-fullyear      = 4DIGIT
            date-month         = 2DIGIT        ;01-12
            date-mday          = 2DIGIT        ;01-28, 01-29, 01-30, 01-31
                                               ;based on month/year
            time               = time-hour time-minute time-second [time-utc]
            time-hour          = 2DIGIT        ;00-23
            time-minute        = 2DIGIT        ;00-59
            time-second        = 2DIGIT        ;00-60
            time-utc           = "Z"

        The following is the representation of the date-time format.

        .. code-block:: text

            YYYYMMDDTHHMMSS

    Description:
        vDatetime is timezone aware and uses a timezone library.
        When a vDatetime object is created from an
        ical string, you can pass a valid timezone identifier. When a
        vDatetime object is created from a Python :py:mod:`datetime` object, it uses the
        tzinfo component, if present. Otherwise a timezone-naive object is
        created. Be aware that there are certain limitations with timezone naive
        DATE-TIME components in the icalendar standard.

    Example:
        The following represents March 2, 2021 at 10:15 AM with local time:

        .. code-block:: pycon

            >>> from icalendar import vDatetime
            >>> datetime = vDatetime.from_ical("20210302T101500")
            >>> datetime.tzname()
            >>> datetime.year
            2021
            >>> datetime.minute
            15

        The following represents March 2, 2021 at 10:15 AM in New York:

        .. code-block:: pycon

            >>> datetime = vDatetime.from_ical("20210302T101500", 'America/New_York')
            >>> datetime.tzname()
            'EST'

        The following represents March 2, 2021 at 10:15 AM in Berlin:

        .. code-block:: pycon

            >>> from zoneinfo import ZoneInfo
            >>> timezone = ZoneInfo("Europe/Berlin")
            >>> vDatetime.from_ical("20210302T101500", timezone)
            datetime.datetime(2021, 3, 2, 10, 15, tzinfo=ZoneInfo(key='Europe/Berlin'))
    """

    default_value: ClassVar[str] = "DATE-TIME"
    params: Parameters

    def __init__(self, dt, /, params: dict[str, Any] | None = None):
        self.dt = dt
        self.params = Parameters(params)
        self.params.update_tzid_from(dt)

    def to_ical(self):
        dt = self.dt

        s = (
            f"{dt.year:04}{dt.month:02}{dt.day:02}"
            f"T{dt.hour:02}{dt.minute:02}{dt.second:02}"
        )
        if self.is_utc():
            s += "Z"
        return s.encode("utf-8")

    @staticmethod
    def from_ical(ical, timezone=None):
        """Create a datetime from the RFC string."""
        tzinfo = None
        if isinstance(timezone, str):
            tzinfo = tzp.timezone(timezone)
        elif timezone is not None:
            tzinfo = timezone

        try:
            timetuple = (
                int(ical[:4]),  # year
                int(ical[4:6]),  # month
                int(ical[6:8]),  # day
                int(ical[9:11]),  # hour
                int(ical[11:13]),  # minute
                int(ical[13:15]),  # second
            )
            if tzinfo:
                return tzp.localize(datetime(*timetuple), tzinfo)
            if not ical[15:]:
                return datetime(*timetuple)
            if ical[15:16] == "Z":
                return tzp.localize_utc(datetime(*timetuple))
        except Exception as e:
            raise ValueError(f"Wrong datetime format: {ical}") from e
        raise ValueError(f"Wrong datetime format: {ical}")

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vDatetime."""
        return [cls(datetime(2025, 11, 10, 16, 52))]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        value = self.dt.strftime("%Y-%m-%dT%H:%M:%S")
        if self.is_utc():
            value += "Z"
        return [name, self.params.to_jcal(exclude_utc=True), self.VALUE.lower(), value]

    def is_utc(self) -> bool:
        """Whether this datetime is UTC."""
        return self.params.is_utc() or is_utc(self.dt)

    @classmethod
    def parse_jcal_value(cls, jcal: str) -> datetime:
        """Parse a jCal string to a :py:class:`datetime.datetime`.

        Raises:
            ~error.JCalParsingError: If it can't parse a date-time value.
        """
        JCalParsingError.validate_value_type(jcal, str, cls)
        utc = jcal.endswith("Z")
        if utc:
            jcal = jcal[:-1]
        try:
            dt = datetime.strptime(jcal, "%Y-%m-%dT%H:%M:%S")
        except ValueError as e:
            raise JCalParsingError("Cannot parse date-time.", cls, value=jcal) from e
        if utc:
            return tzp.localize_utc(dt)
        return dt

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        params = Parameters.from_jcal_property(jcal_property)
        with JCalParsingError.reraise_with_path_added(3):
            dt = cls.parse_jcal_value(jcal_property[3])
        if params.tzid:
            dt = tzp.localize(dt, params.tzid)
        return cls(
            dt,
            params=params,
        )


__all__ = ["vDatetime"]



"""DATE property type from :rfc:`5545`."""

from datetime import date, datetime
from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters

from .base import TimeBase


class vDate(TimeBase):
    """Date

    Value Name:
        DATE

    Purpose:
        This value type is used to identify values that contain a
        calendar date.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            date               = date-value

            date-value         = date-fullyear date-month date-mday
            date-fullyear      = 4DIGIT
            date-month         = 2DIGIT        ;01-12
            date-mday          = 2DIGIT        ;01-28, 01-29, 01-30, 01-31
                                               ;based on month/year

    Description:
        If the property permits, multiple "date" values are
        specified as a COMMA-separated list of values.  The format for the
        value type is based on the [ISO.8601.2004] complete
        representation, basic format for a calendar date.  The textual
        format specifies a four-digit year, two-digit month, and two-digit
        day of the month.  There are no separator characters between the
        year, month, and day component text.

    Example:
        The following represents July 14, 1997:

        .. code-block:: ics

            19970714

        .. code-block:: pycon

            >>> from icalendar.prop import vDate
            >>> date = vDate.from_ical('19970714')
            >>> date.year
            1997
            >>> date.month
            7
            >>> date.day
            14
    """

    default_value: ClassVar[str] = "DATE"
    params: Parameters

    def __init__(self, dt, params: dict[str, Any] | None = None):
        if not isinstance(dt, date):
            raise TypeError("Value MUST be a date instance")
        self.dt = dt
        self.params = Parameters(params or {})

    def to_ical(self):
        s = f"{self.dt.year:04}{self.dt.month:02}{self.dt.day:02}"
        return s.encode("utf-8")

    @staticmethod
    def from_ical(ical):
        try:
            timetuple = (
                int(ical[:4]),  # year
                int(ical[4:6]),  # month
                int(ical[6:8]),  # day
            )
            return date(*timetuple)
        except Exception as e:
            raise ValueError(f"Wrong date format {ical}") from e

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vDate."""
        return [cls(date(2025, 11, 10))]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        return [
            name,
            self.params.to_jcal(),
            self.VALUE.lower(),
            self.dt.strftime("%Y-%m-%d"),
        ]

    @classmethod
    def parse_jcal_value(cls, jcal: str) -> date:
        """Parse a jCal string to a :py:class:`datetime.date`.

        Raises:
            ~error.JCalParsingError: If it can't parse a date.
        """
        JCalParsingError.validate_value_type(jcal, str, cls)
        try:
            return datetime.strptime(jcal, "%Y-%m-%d").date()
        except ValueError as e:
            raise JCalParsingError("Cannot parse date.", cls, value=jcal) from e

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        with JCalParsingError.reraise_with_path_added(3):
            value = cls.parse_jcal_value(jcal_property[3])
        return cls(
            value,
            params=Parameters.from_jcal_property(jcal_property),
        )


__all__ = ["vDate"]



"""BYMONTH value type of RECUR from :rfc:`5545` and :rfc:`7529`."""

from typing import Any

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters


class vMonth(int):
    """The number of the month for recurrence.

    In :rfc:`5545`, this is just an int.
    In :rfc:`7529`, this can be followed by `L` to indicate a leap month.

    .. code-block:: pycon

        >>> from icalendar import vMonth
        >>> vMonth(1) # first month January
        vMonth('1')
        >>> vMonth("5L") # leap month in Hebrew calendar
        vMonth('5L')
        >>> vMonth(1).leap
        False
        >>> vMonth("5L").leap
        True

    Definition from RFC:

    .. code-block:: text

        type-bymonth = element bymonth {
           xsd:positiveInteger |
           xsd:string
        }
    """

    params: Parameters

    def __new__(cls, month: str | int, /, params: dict[str, Any] | None = None):
        if isinstance(month, vMonth):
            return cls(month.to_ical().decode())
        if isinstance(month, str):
            if month.isdigit():
                month_index = int(month)
                leap = False
            else:
                if not month or month[-1] != "L" or not month[:-1].isdigit():
                    raise ValueError(f"Invalid month: {month!r}")
                month_index = int(month[:-1])
                leap = True
        else:
            leap = False
            month_index = int(month)
        self = super().__new__(cls, month_index)
        self.leap = leap
        self.params = Parameters(params)
        return self

    def to_ical(self) -> bytes:
        """The ical representation."""
        return str(self).encode("utf-8")

    @classmethod
    def from_ical(cls, ical: str):
        return cls(ical)

    @property
    def leap(self) -> bool:
        """Whether this is a leap month."""
        return self._leap

    @leap.setter
    def leap(self, value: bool) -> None:
        self._leap = value

    def __repr__(self) -> str:
        """repr(self)"""
        return f"{self.__class__.__name__}({str(self)!r})"

    def __str__(self) -> str:
        """str(self)"""
        return f"{int(self)}{'L' if self.leap else ''}"

    @classmethod
    def parse_jcal_value(cls, value: Any) -> Self:
        """Parse a jCal value for vMonth.

        Raises:
            ~error.JCalParsingError: If the value is not a valid month.
        """
        JCalParsingError.validate_value_type(value, (str, int), cls)
        try:
            return cls(value)
        except ValueError as e:
            raise JCalParsingError(
                "The value must be a string or an integer.", cls, value=value
            ) from e


__all__ = ["vMonth"]



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, TypeVar

from icalendar.parser_tools import to_unicode

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

KT = TypeVar("KT")
VT = TypeVar("VT")


def canonsort_keys(
    keys: Iterable[KT], canonical_order: Iterable[KT] | None = None
) -> list[KT]:
    """Sort leading keys according to a canonical order.

    Keys specified in ``canonical_order`` appear first in that order.
    Remaining keys appear alphabetically at the end.

    Parameters:
        keys: The keys to sort.
        canonical_order: The preferred order for leading keys.
            Keys not in this sequence are sorted alphabetically after
            the canonical ones. If ``None``, all keys are sorted
            alphabetically.

    Returns:
        A new list of keys sorted by canonical order first, then alphabetically.

    Example:
        ..  code-block:: pycon

            >>> from icalendar.caselessdict import canonsort_keys
            >>> canonsort_keys(["C", "A", "B"], ["B", "C"])
            ['B', 'C', 'A']
    """
    canonical_map = {k: i for i, k in enumerate(canonical_order or [])}
    head = [k for k in keys if k in canonical_map]
    tail = [k for k in keys if k not in canonical_map]
    return sorted(head, key=lambda k: canonical_map[k]) + sorted(tail)


def canonsort_items(
    dict1: Mapping[KT, VT], canonical_order: Iterable[KT] | None = None
) -> list[tuple[KT, VT]]:
    """Sort items from a mapping according to a canonical key order.

    Parameters:
        dict1: The mapping whose items to sort.
        canonical_order: The preferred order for leading keys.
            If ``None``, all keys are sorted alphabetically.

    Returns:
        A list of ``(key, value)`` tuples sorted by canonical order.

    Example:
        ..  code-block:: pycon

            >>> from icalendar.caselessdict import canonsort_items
            >>> canonsort_items({"C": 3, "A": 1, "B": 2}, ["B", "C"])
            [('B', 2), ('C', 3), ('A', 1)]
    """
    return [(k, dict1[k]) for k in canonsort_keys(dict1.keys(), canonical_order)]


class CaselessDict(OrderedDict):
    """A case-insensitive dictionary that uses strings as keys.

    All keys are stored in uppercase internally, but values retain
    their original case. Keys can be provided as ``str`` or ``bytes``.
    They are converted to Unicode via :func:`~icalendar.parser_tools.to_unicode`,
    then uppercased before storage.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Parameters:
            *args: Positional arguments passed to :class:`~collections.OrderedDict`.
            **kwargs: Keyword arguments passed to :class:`~collections.OrderedDict`.

        Example:

            Create a new ``CaselessDict`` and normalize existing keys to uppercase.

            ..  code-block:: pycon

                >>> from icalendar.caselessdict import CaselessDict
                >>> d = CaselessDict(summary="Meeting")
                >>> d["SUMMARY"]
                'Meeting'
                >>> "summary" in d
                True
        """
        super().__init__(*args, **kwargs)
        for key, value in self.items():
            key_upper = to_unicode(key).upper()
            if key != key_upper:
                super().__delitem__(key)
                self[key_upper] = value

    __hash__ = None

    def __getitem__(self, key: Any) -> Any:
        """Get the item from the ``CaselessDict`` instance by
        ``key``, case-insensitively.

        Parameters:
            key: The key to look up, case-insensitively.

        Returns:
            The (key, value) pair associated with the uppercased key.

        Raises:
            KeyError: If the key is not found.
        """
        key = to_unicode(key)
        return super().__getitem__(key.upper())

    def __setitem__(self, key: Any, value: Any) -> None:
        """Set a (key, value) pair, storing the key in uppercase.

        Parameters:
            key: The key of the pair, case-insensitive.
            value: The value to associate with the key.
        """
        key = to_unicode(key)
        super().__setitem__(key.upper(), value)

    def __delitem__(self, key: Any) -> None:
        """Delete a (key, value) pair by its case-insensitive key.

        Parameters:
            key: The key to delete, case-insensitively.

        Raises:
            KeyError: If the key is not found.
        """
        key = to_unicode(key)
        super().__delitem__(key.upper())

    def __contains__(self, key: Any) -> bool:
        """Check whether a key exists in the mapping, case-insensitively.

        Parameters:
            key: The key to check case-insensitively.

        Returns:
            ``True`` if the uppercased key exists, else ``False``.
        """
        key = to_unicode(key)
        return super().__contains__(key.upper())

    def get(self, key: Any, default: Any = None) -> Any:
        """Return the ``key``, optionally with a ``default`` value.

        Parameters:
            key: The key to look up, case-insensitively.
            default: The value to return if the key is not found.

        Returns:
            The value for the key, if present, else the value specified by ``default``.
        """
        key = to_unicode(key)
        return super().get(key.upper(), default)

    def setdefault(self, key: Any, value: Any = None) -> Any:
        """Create the (key, value) pair, optionally with a ``value``.

        Once set, to change default value use :meth:`update`.

        Parameters:
            key: The key to look up or create, case-insensitively.
            value: The default value to set, if given, else ``None``.

        Returns:
            The value for the key.
        """
        key = to_unicode(key)
        return super().setdefault(key.upper(), value)

    def pop(self, key: Any, default: Any = None) -> Any:
        """Remove and return the value for ``key``, or ``default`` if not found.

        Parameters:
            key: The key to remove, case-insensitively.
            default: The value to return if the key is not found.

        Returns:
            The removed value, or the value of ``default``.
        """
        key = to_unicode(key)
        return super().pop(key.upper(), default)

    def popitem(self) -> tuple[Any, Any]:
        """Remove and return the last inserted (key, value) pair.

        Returns:
            A (key, value) tuple.

        Raises:
            KeyError: If the dictionary is empty.
        """
        return super().popitem()

    def has_key(self, key: Any) -> bool:
        """Check whether a key exists, case-insensitively.

        This is a legacy method. Use ``key in dict`` instead.

        Parameters:
            key: The key to check, case-insensitively.

        Returns:
            ``True`` if the key exists, else ``False``.
        """
        key = to_unicode(key)
        return super().__contains__(key.upper())

    def update(self, *args: Any, **kwargs: Any) -> None:
        """Update the dictionary with (key, value) pairs, normalizing keys to uppercase.

        Multiple keys that differ only in case will overwrite each other.
        Only the last value is retained.

        Parameters:
            *args: Mappings or iterables of (key, value) pairs.
            **kwargs: Additional (key, value) pairs.
        """
        # Multiple keys where key1.upper() == key2.upper() will be lost.
        mappings = list(args) + [kwargs]
        for mapping in mappings:
            if hasattr(mapping, "items"):
                mapping = iter(mapping.items())  # noqa: PLW2901
            for key, value in mapping:
                self[key] = value

    def copy(self) -> Self:
        """Return a shallow copy of the dictionary.

        Returns:
            A new instance of the same type with the same contents.
        """
        return type(self)(super().copy())

    def __repr__(self) -> str:
        """Return a string representation of the dictionary.

        Returns:
            A string in the form ``CaselessDict({...})``.
        """
        return f"{type(self).__name__}({dict(self)})"

    def __eq__(self, other: object) -> bool:
        """Check equality with another dictionary.

        Two ``CaselessDict`` instances are equal if they contain the same
        (key, value) pairs after uppercasing keys. Comparison with a regular
        ``dict`` also works.

        Parameters:
            other: The object to compare.

        Returns:
            ``True`` if equal, ``NotImplemented`` if ``other`` is not a ``dict``.
        """
        if not isinstance(other, dict):
            return NotImplemented
        return self is other or dict(self.items()) == dict(other.items())

    def __ne__(self, other: object) -> bool:
        """Check inequality with another dictionary.

        Parameters:
            other: The object to compare.

        Returns:
            ``True`` if not equal, else ``False``.
        """
        return not self == other

    # A list of keys that must appear first in sorted_keys and sorted_items;
    # must be uppercase.
    canonical_order = None

    def sorted_keys(self) -> list[str]:
        """Sort keys according to the canonical order for this class.

        Keys listed in :attr:`canonical_order` appear first in that order.
        Remaining keys appear alphabetically at the end.

        Returns:
            A sorted list of keys.
        """
        return canonsort_keys(self.keys(), self.canonical_order)

    def sorted_items(self) -> list[tuple[Any, Any]]:
        """Sort items according to the canonical order for this class.

        Items whose keys are listed in :attr:`canonical_order` appear first
        in that order. Remaining items appear alphabetically by key.

        Returns:
            A sorted list of (key, value) tuples.
        """
        return canonsort_items(self, self.canonical_order)


__all__ = ["CaselessDict", "canonsort_items", "canonsort_keys"]



"""TIME property type from :rfc:`5545`."""

import re
from datetime import datetime, time, timezone, tzinfo
from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.timezone import tzp
from icalendar.timezone.tzid import is_utc

from .base import TimeBase

TIME_JCAL_REGEX = re.compile(
    r"^(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):(?P<second>[0-9]{2})(?P<utc>Z)?$"
)


class vTime(TimeBase):
    """Time

    Value Name:
        TIME

    Purpose:
        This value type is used to identify values that contain a
        time of day.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            time         = time-hour time-minute time-second [time-utc]

            time-hour    = 2DIGIT        ;00-23
            time-minute  = 2DIGIT        ;00-59
            time-second  = 2DIGIT        ;00-60
            ;The "60" value is used to account for positive "leap" seconds.

            time-utc     = "Z"

    Description:
        If the property permits, multiple "time" values are
        specified by a COMMA-separated list of values.  No additional
        content value encoding (i.e., BACKSLASH character encoding, see
        vText) is defined for this value type.

        The "TIME" value type is used to identify values that contain a
        time of day.  The format is based on the [ISO.8601.2004] complete
        representation, basic format for a time of day.  The text format
        consists of a two-digit, 24-hour of the day (i.e., values 00-23),
        two-digit minute in the hour (i.e., values 00-59), and two-digit
        seconds in the minute (i.e., values 00-60).  The seconds value of
        60 MUST only be used to account for positive "leap" seconds.
        Fractions of a second are not supported by this format.

        In parallel to the "DATE-TIME" definition above, the "TIME" value
        type expresses time values in three forms:

        The form of time with UTC offset MUST NOT be used.  For example,
        the following is not valid for a time value:

        .. code-block:: ics

            230000-0800        ;Invalid time format

        **FORM #1 LOCAL TIME**

        The local time form is simply a time value that does not contain
        the UTC designator nor does it reference a time zone.  For
        example, 11:00 PM:

        .. code-block:: ics

            230000

        Time values of this type are said to be "floating" and are not
        bound to any time zone in particular.  They are used to represent
        the same hour, minute, and second value regardless of which time
        zone is currently being observed.  For example, an event can be
        defined that indicates that an individual will be busy from 11:00
        AM to 1:00 PM every day, no matter which time zone the person is
        in.  In these cases, a local time can be specified.  The recipient
        of an iCalendar object with a property value consisting of a local
        time, without any relative time zone information, SHOULD interpret
        the value as being fixed to whatever time zone the "ATTENDEE" is
        in at any given moment.  This means that two "Attendees", may
        participate in the same event at different UTC times; floating
        time SHOULD only be used where that is reasonable behavior.

        In most cases, a fixed time is desired.  To properly communicate a
        fixed time in a property value, either UTC time or local time with
        time zone reference MUST be specified.

        The use of local time in a TIME value without the "TZID" property
        parameter is to be interpreted as floating time, regardless of the
        existence of "VTIMEZONE" calendar components in the iCalendar
        object.

        **FORM #2: UTC TIME**

        UTC time, or absolute time, is identified by a LATIN CAPITAL
        LETTER Z suffix character, the UTC designator, appended to the
        time value.  For example, the following represents 07:00 AM UTC:

        .. code-block:: ics

            070000Z

        The "TZID" property parameter MUST NOT be applied to TIME
        properties whose time values are specified in UTC.

        **FORM #3: LOCAL TIME AND TIME ZONE REFERENCE**

        The local time with reference to time zone information form is
        identified by the use the "TZID" property parameter to reference
        the appropriate time zone definition.

        Example:
            The following represents 8:30 AM in New York in winter,
            five hours behind UTC, in each of the three formats:

        .. code-block:: ics

            083000
            133000Z
            TZID=America/New_York:083000
    """

    default_value: ClassVar[str] = "TIME"
    params: Parameters

    def __init__(self, *args, params: dict[str, Any] | None = None):
        if len(args) == 1:
            if not isinstance(args[0], (time, datetime)):
                raise ValueError(f"Expected a datetime.time, got: {args[0]}")
            self.dt = args[0]
        else:
            self.dt = time(*args)
        self.params = Parameters(params or {})
        self.params.update_tzid_from(self.dt)

    def to_ical(self):
        value = self.dt.strftime("%H%M%S")
        if self.is_utc():
            value += "Z"
        return value

    def is_utc(self) -> bool:
        """Whether this time is UTC."""
        return self.params.is_utc() or is_utc(self.dt)

    @staticmethod
    def from_ical(ical: str, timezone: str | None | tzinfo = None) -> time:
        """Convert an ical string into a time.

        This method supports parsing the three forms of time values defined in :rfc:`5545#section-3.3.12`:
            - Local time (floating)
            - UTC time
            - Local time with time zone reference

        Returns:
            A :class:`datetime.time` object representing the parsed time, with timezone information if applicable.

        Raises:
            ValueError: if the provided string cannot be parsed as a time.
        """
        tzinfo = None
        if isinstance(timezone, str):
            tzinfo = tzp.timezone(timezone)
        elif timezone is not None:
            tzinfo = timezone

        try:
            if isinstance(ical, bytes):
                ical = ical.decode()
            utc = ical.endswith("Z")
            if utc:
                ical = ical[:-1]
            timetuple = (int(ical[:2]), int(ical[2:4]), int(ical[4:6]))
            if tzinfo:
                return tzp.localize(time(*timetuple), tzinfo)
            if utc:
                return tzp.localize_utc(time(*timetuple))
            return time(*timetuple)
        except Exception as e:
            raise ValueError(f"Expected time, got: {ical}") from e

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vTime."""
        return [cls(time(12, 30))]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        value = self.dt.strftime("%H:%M:%S")
        if self.is_utc():
            value += "Z"
        return [name, self.params.to_jcal(exclude_utc=True), self.VALUE.lower(), value]

    @classmethod
    def parse_jcal_value(cls, jcal: str) -> time:
        """Parse a jCal string to a :py:class:`datetime.time`.

        Raises:
            ~error.JCalParsingError: If it can't parse a time.
        """
        JCalParsingError.validate_value_type(jcal, str, cls)
        match = TIME_JCAL_REGEX.match(jcal)
        if match is None:
            raise JCalParsingError("Cannot parse time.", cls, value=jcal)
        hour = int(match.group("hour"))
        minute = int(match.group("minute"))
        second = int(match.group("second"))
        utc = bool(match.group("utc"))
        return time(hour, minute, second, tzinfo=timezone.utc if utc else None)

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        with JCalParsingError.reraise_with_path_added(3):
            value = cls.parse_jcal_value(jcal_property[3])
        return cls(
            value,
            params=Parameters.from_jcal_property(jcal_property),
        )


__all__ = ["vTime"]



"""Functions for parsing parameters."""

from __future__ import annotations

import functools
import os
import re
from datetime import datetime, time
from typing import TYPE_CHECKING, Any, Protocol

from icalendar.caselessdict import CaselessDict
from icalendar.error import JCalParsingError
from icalendar.parser.string import validate_token
from icalendar.parser_tools import (
    DEFAULT_ENCODING,
    SEQUENCE_TYPES,
)
from icalendar.timezone.tzid import tzid_from_dt

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from icalendar.enums import VALUE
    from icalendar.prop import VPROPERTY


class HasToIcal(Protocol):
    """Protocol for objects with a to_ical method."""

    def to_ical(self) -> bytes:
        """Convert to iCalendar format."""
        ...


def param_value(
    value: Sequence[str] | str | HasToIcal, always_quote: bool = False
) -> str:
    """Convert a parameter value to its iCalendar representation.

    Applies :rfc:`6868` escaping and optionally quotes the value according
    to :rfc:`5545` parameter value formatting rules.

    Parameters:
        value: The parameter value to convert. Can be a sequence, string, or
            object with a ``to_ical()`` method.
        always_quote: If ``True``, always enclose the value in double quotes.
            Defaults to ``False`` (only quote when necessary).

    Returns:
        The formatted parameter value, escaped and quoted as needed.
    """
    if isinstance(value, SEQUENCE_TYPES):
        return q_join(map(rfc_6868_escape, value), always_quote=always_quote)
    if isinstance(value, str):
        return dquote(rfc_6868_escape(value), always_quote=always_quote)
    return dquote(rfc_6868_escape(value.to_ical().decode(DEFAULT_ENCODING)))


# Could be improved


UNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7f",:;]')
QUNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7f"]')


def validate_param_value(value: str, quoted: bool = True) -> None:
    """Validate a parameter value for unsafe characters.

    Checks parameter values for characters that are not allowed according to
    :rfc:`5545`. Uses different validation rules for quoted and unquoted values.

    Parameters:
        value: The parameter value to validate.
        quoted: If ``True``, validate as a quoted value (allows more characters).
            If ``False``, validate as an unquoted value (stricter).
            Defaults to ``True``.

    Raises:
        ValueError: If the value contains unsafe characters for its quote state.
    """
    validator = QUNSAFE_CHAR if quoted else UNSAFE_CHAR
    if validator.findall(value):
        raise ValueError(value)


# chars presence of which in parameter value will be cause the value
# to be enclosed in double-quotes
QUOTABLE = re.compile("[,;:’]")  # noqa: RUF001


def dquote(val: str, always_quote: bool = False) -> str:
    """Enclose parameter values in double quotes when needed.

    Parameter values containing special characters ``,``, ``;``,
    ``:`` or ``'`` must be enclosed
    in double quotes according to :rfc:`5545`. Double-quote characters in the
    value are replaced with single quotes since they're forbidden in parameter
    values.

    Parameters:
        val: The parameter value to quote.
        always_quote: If ``True``, always enclose in quotes regardless of content.
            Defaults to ``False`` (only quote when necessary).

    Returns:
        The value, enclosed in double quotes if needed or requested.
    """
    # a double-quote character is forbidden to appear in a parameter value
    # so replace it with a single-quote character
    val = val.replace('"', "'")
    if QUOTABLE.search(val) or always_quote:
        return f'"{val}"'
    return val


# parsing helper
def q_split(st: str, sep: str = ",", maxsplit: int = -1) -> list[str]:
    """Split a string on a separator, respecting double quotes.

    Splits the string on the separator character, but ignores separators that
    appear inside double-quoted sections. This is needed for parsing parameter
    values that may contain quoted strings.

    Parameters:
        st: The string to split.
        sep: The separator character. Defaults to ``,``.
        maxsplit: Maximum number of splits to perform. If ``-1`` (default),
            then perform all possible splits.

    Returns:
        The split string parts.

    Examples:
        .. code-block:: pycon

            >>> from icalendar.parser import q_split
            >>> q_split('a,b,c')
            ['a', 'b', 'c']
            >>> q_split('a,"b,c",d')
            ['a', '"b,c"', 'd']
            >>> q_split('a;b;c', sep=';')
            ['a', 'b', 'c']
    """
    if maxsplit == 0:
        return [st]

    result = []
    cursor = 0
    length = len(st)
    inquote = 0
    splits = 0
    for i, ch in enumerate(st):
        if ch == '"':
            inquote = not inquote
        if not inquote and ch == sep:
            result.append(st[cursor:i])
            cursor = i + 1
            splits += 1
        if i + 1 == length or splits == maxsplit:
            result.append(st[cursor:])
            break
    return result


def q_join(lst: Sequence[str], sep: str = ",", always_quote: bool = False) -> str:
    """Join a list with a separator, quoting items as needed.

    Joins list items with the separator, applying :func:`dquote` to each item
    to add double quotes when they contain special characters.

    Parameters:
        lst: The list of items to join.
        sep: The separator to use. Defaults to ``,``.
        always_quote: If ``True``, always quote all items. Defaults to ``False``
            (only quote when necessary).

    Returns:
        The joined string with items quoted as needed.

    Examples:
        .. code-block:: pycon

            >>> from icalendar.parser import q_join
            >>> q_join(['a', 'b', 'c'])
            'a,b,c'
            >>> q_join(['plain', 'has,comma'])
            'plain,"has,comma"'
    """
    return sep.join(dquote(itm, always_quote=always_quote) for itm in lst)


def single_string_parameter(func: Callable | None = None, upper=False):
    """Create a parameter getter/setter for a single string parameter.

    Parameters:
        upper: Convert the value to uppercase
        func: The function to decorate.

    Returns:
        The property for the parameter or a decorator for the parameter
        if func is ``None``.
    """

    def decorator(func):
        name = func.__name__

        @functools.wraps(func)
        def fget(self: Parameters):
            """Get the value."""
            value = self.get(name)
            if value is not None and upper:
                value = value.upper()
            return value

        def fset(self: Parameters, value: str | None):
            """Set the value"""
            if value is None:
                fdel(self)
            else:
                if upper:
                    value = value.upper()
                self[name] = value

        def fdel(self: Parameters):
            """Delete the value."""
            self.pop(name, None)

        return property(fget, fset, fdel, doc=func.__doc__)

    if func is None:
        return decorator
    return decorator(func)


class Parameters(CaselessDict):
    """Parser and generator of Property parameter strings.

    It knows nothing of datatypes.
    Its main concern is textual structure.

    Examples:

        Modify parameters:

        .. code-block:: pycon

            >>> from icalendar import Parameters
            >>> params = Parameters()
            >>> params['VALUE'] = 'TEXT'
            >>> params.value
            'TEXT'
            >>> params
            Parameters({'VALUE': 'TEXT'})

        Create new parameters:

        .. code-block:: pycon

            >>> params = Parameters(value="BINARY")
            >>> params.value
            'BINARY'

        Set a default:

        .. code-block:: pycon

            >>> params = Parameters(value="BINARY", default_value="TEXT")
            >>> params
            Parameters({'VALUE': 'BINARY'})

    """

    def __init__(self, *args, **kwargs):
        """Create new parameters."""
        if args and args[0] is None:
            # allow passing None
            args = args[1:]
        defaults = {
            key[8:]: kwargs.pop(key)
            for key in list(kwargs.keys())
            if key.lower().startswith("default_")
        }
        super().__init__(*args, **kwargs)
        for key, value in defaults.items():
            self.setdefault(key, value)

    # The following paremeters must always be enclosed in double quotes
    always_quoted = (
        "ALTREP",
        "DELEGATED-FROM",
        "DELEGATED-TO",
        "DIR",
        "MEMBER",
        "SENT-BY",
        # Part of X-APPLE-STRUCTURED-LOCATION
        "X-ADDRESS",
        "X-TITLE",
        # RFC 9253
        "LINKREL",
    )
    # this is quoted should one of the values be present
    quote_also = {
        # This is escaped in the RFC
        "CN": " '",
    }

    def params(self):
        """In RFC 5545 keys are called parameters, so this is to be consitent
        with the naming conventions.
        """
        return self.keys()

    def to_ical(self, sorted: bool = True):  # noqa: A002
        """Returns an :rfc:`5545` representation of the parameters.

        Parameters:
            sorted (bool): Sort the parameters before encoding.
            exclude_utc (bool): Exclude TZID if it is set to ``"UTC"``
        """
        result = []
        items = list(self.items())
        if sorted:
            items.sort()

        for key, value in items:
            if key == "TZID" and value == "UTC":
                # The "TZID" property parameter MUST NOT be applied to DATE-TIME
                # properties whose time values are specified in UTC.
                continue
            upper_key = key.upper()
            check_quoteable_characters = self.quote_also.get(key.upper())
            always_quote = upper_key in self.always_quoted or (
                check_quoteable_characters
                and any(c in value for c in check_quoteable_characters)
            )
            quoted_value = param_value(value, always_quote=always_quote)
            if isinstance(quoted_value, str):
                quoted_value = quoted_value.encode(DEFAULT_ENCODING)
            # CaselessDict keys are always unicode
            result.append(upper_key.encode(DEFAULT_ENCODING) + b"=" + quoted_value)
        return b";".join(result)

    @classmethod
    def from_ical(cls, st, strict=False):
        """Parses the parameter format from ical text format."""

        # parse into strings
        result = cls()
        for param in q_split(st, ";"):
            try:
                key, val = q_split(param, "=", maxsplit=1)
                validate_token(key)
                # Property parameter values that are not in quoted
                # strings are case insensitive.
                vals = []
                for v in q_split(val, ","):
                    if v.startswith('"') and v.endswith('"'):
                        v2 = v.strip('"')
                        validate_param_value(v2, quoted=True)
                        vals.append(rfc_6868_unescape(v2))
                    else:
                        validate_param_value(v, quoted=False)
                        if strict:
                            vals.append(rfc_6868_unescape(v.upper()))
                        else:
                            vals.append(rfc_6868_unescape(v))
                if not vals:
                    result[key] = val
                elif len(vals) == 1:
                    result[key] = vals[0]
                else:
                    result[key] = vals
            except ValueError as exc:  # noqa: PERF203
                raise ValueError(
                    f"{param!r} is not a valid parameter string: {exc}"
                ) from exc
        return result

    @single_string_parameter(upper=True)
    def value(self) -> VALUE | str | None:
        """The VALUE parameter from :rfc:`5545`.

        Description:
            This parameter specifies the value type and format of
            the property value.  The property values MUST be of a single value
            type.  For example, a "RDATE" property cannot have a combination
            of DATE-TIME and TIME value types.

            If the property's value is the default value type, then this
            parameter need not be specified.  However, if the property's
            default value type is overridden by some other allowable value
            type, then this parameter MUST be specified.

            Applications MUST preserve the value data for x-name and iana-
            token values that they don't recognize without attempting to
            interpret or parse the value data.

        For convenience, using this property, the value will be converted to
        an uppercase string.

        .. code-block:: pycon

            >>> from icalendar import Parameters
            >>> params = Parameters()
            >>> params.value = "unknown"
            >>> params
            Parameters({'VALUE': 'UNKNOWN'})

        """

    def _parameter_value_to_jcal(
        self, value: str | float | list | VPROPERTY
    ) -> str | int | float | list[str] | list[int] | list[float]:
        """Convert a parameter value to jCal format.

        Parameters:
            value: The parameter value

        Returns:
            The jCal representation of the parameter value
        """
        if isinstance(value, list):
            return [self._parameter_value_to_jcal(v) for v in value]
        if hasattr(value, "to_jcal"):
            # proprty values respond to this
            jcal = value.to_jcal()
            # we only need the value part
            if len(jcal) == 4:
                return jcal[3]
            return jcal[3:]
        for t in (int, float, str):
            if isinstance(value, t):
                return t(value)
        raise TypeError(
            "Unsupported parameter value type for jCal conversion: "
            f"{type(value)} {value!r}"
        )

    def to_jcal(self, exclude_utc=False) -> dict[str, str]:
        """Return the jCal representation of the parameters.

        Parameters:
            exclude_utc (bool): Exclude the TZID parameter if it is UTC
        """
        jcal = {
            k.lower(): self._parameter_value_to_jcal(v)
            for k, v in self.items()
            if k.lower() != "value"
        }
        if exclude_utc and jcal.get("tzid") == "UTC":
            del jcal["tzid"]
        return jcal

    @single_string_parameter
    def tzid(self) -> str | None:
        """The TZID parameter from :rfc:`5545`."""

    def is_utc(self):
        """Whether the TZID parameter is UTC."""
        return self.tzid == "UTC"

    def update_tzid_from(self, dt: datetime | time | Any) -> None:
        """Update the TZID parameter from a datetime object.

        This sets the TZID parameter or deletes it according to the datetime.
        :rfc:`5545#section-3.2.19` prohibits TZID on UTC datetimes,
        which use the ``Z`` suffix instead.
        """
        if isinstance(dt, (datetime, time)):
            tzid = tzid_from_dt(dt)
            if tzid != "UTC":
                # UTC uses Z suffix and does not appear as TZID parameter
                self.tzid = tzid

    @classmethod
    def from_jcal(cls, jcal: dict[str : str | list[str]]):
        """Parse jCal parameters."""
        if not isinstance(jcal, dict):
            raise JCalParsingError("The parameters must be a mapping.", cls)
        for name, value in jcal.items():
            if not isinstance(name, str):
                raise JCalParsingError(
                    "All parameter names must be strings.", cls, value=name
                )
            if not (
                (
                    isinstance(value, list)
                    and all(isinstance(v, (str, int, float)) for v in value)
                    and value
                )
                or isinstance(value, (str, int, float))
            ):
                raise JCalParsingError(
                    "Parameter values must be a string, integer or "
                    "float or a list of those.",
                    cls,
                    name,
                    value=value,
                )
        return cls(jcal)

    @classmethod
    def from_jcal_property(cls, jcal_property: list):
        """Create the parameters for a jCal property.

        Parameters:
            jcal_property (list): The jCal property [name, params, value, ...]
            default_value (str, optional): The default value of the property.
                If this is given, the default value will not be set.
        """
        if not isinstance(jcal_property, list) or len(jcal_property) < 4:
            raise JCalParsingError(
                "The property must be a list with at least 4 items.", cls
            )
        jcal_params = jcal_property[1]
        with JCalParsingError.reraise_with_path_added(1):
            self = cls.from_jcal(jcal_params)
        if self.is_utc():
            del self.tzid  # we do not want this parameter
        return self


RFC_6868_UNESCAPE_REGEX = re.compile(r"\^\^|\^n|\^'")


def rfc_6868_unescape(param_value: str) -> str:
    """Take care of :rfc:`6868` unescaping.

    - ^^ -> ^
    - ^n -> system specific newline
    - ^' -> "
    - ^ with others stay intact
    """
    replacements = {
        "^^": "^",
        "^n": os.linesep,
        "^'": '"',
    }
    return RFC_6868_UNESCAPE_REGEX.sub(
        lambda m: replacements.get(m.group(0), m.group(0)), param_value
    )


RFC_6868_ESCAPE_REGEX = re.compile(r'\^|\r\n|\r|\n|"')


def rfc_6868_escape(param_value: str) -> str:
    """Take care of :rfc:`6868` escaping.

    - ^ -> ^^
    - " -> ^'
    - newline -> ^n
    """
    replacements = {
        "^": "^^",
        "\n": "^n",
        "\r": "^n",
        "\r\n": "^n",
        '"': "^'",
    }
    return RFC_6868_ESCAPE_REGEX.sub(
        lambda m: replacements.get(m.group(0), m.group(0)), param_value
    )


__all__ = [
    "Parameters",
    "dquote",
    "param_value",
    "q_join",
    "q_split",
    "rfc_6868_escape",
    "rfc_6868_unescape",
    "validate_param_value",
]



"""Special parsing for calendar components."""

from .component import ComponentIcalParser


class CalendarIcalParser(ComponentIcalParser):
    """A parser for calendar components."""

    def prepare_components(self) -> None:
        """Prepare the parsed components.

        This handles timezone forward references.
        """
        all_timezones_so_far = True
        for comp in self._components:
            for component in comp.subcomponents:
                if component.name == "VTIMEZONE":
                    if not all_timezones_so_far:
                        # If a preceding component refers to a VTIMEZONE
                        # defined later in the source string (forward
                        # references are allowed by RFC 5545), then the
                        # earlier component may have the wrong timezone
                        # attached.
                        # However, during computation of comps, all VTIMEZONEs
                        # observed do end up in the timezone cache. So simply
                        # re-running from_ical will rely on the cache for those
                        # forward references to produce the correct result.
                        # See test_create_america_new_york_forward_reference.
                        self.initialize_parsing()
                        self.parse_content_lines()
                        return
                else:
                    all_timezones_so_far = False


__all__ = ["CalendarIcalParser"]



"""DURATION property type from :rfc:`5545`."""

import re
from datetime import timedelta
from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import InvalidCalendar, JCalParsingError
from icalendar.parser import Parameters

from .base import TimeBase

DURATION_REGEX = re.compile(
    r"([-+]?)P(?:(\d+)W)?(?:(\d+)D)?(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+)S)?)?$"
)


class vDuration(TimeBase):
    """Duration

    Value Name:
        DURATION

    Purpose:
        This value type is used to identify properties that contain
        a duration of time.

    Format Definition:
        This value type is defined by the following notation:

        .. code-block:: text

            dur-value  = (["+"] / "-") "P" (dur-date / dur-time / dur-week)

            dur-date   = dur-day [dur-time]
            dur-time   = "T" (dur-hour / dur-minute / dur-second)
            dur-week   = 1*DIGIT "W"
            dur-hour   = 1*DIGIT "H" [dur-minute]
            dur-minute = 1*DIGIT "M" [dur-second]
            dur-second = 1*DIGIT "S"
            dur-day    = 1*DIGIT "D"

    Description:
        If the property permits, multiple "duration" values are
        specified by a COMMA-separated list of values.  The format is
        based on the [ISO.8601.2004] complete representation basic format
        with designators for the duration of time.  The format can
        represent nominal durations (weeks and days) and accurate
        durations (hours, minutes, and seconds).  Note that unlike
        [ISO.8601.2004], this value type doesn't support the "Y" and "M"
        designators to specify durations in terms of years and months.
        The duration of a week or a day depends on its position in the
        calendar.  In the case of discontinuities in the time scale, such
        as the change from standard time to daylight time and back, the
        computation of the exact duration requires the subtraction or
        addition of the change of duration of the discontinuity.  Leap
        seconds MUST NOT be considered when computing an exact duration.
        When computing an exact duration, the greatest order time
        components MUST be added first, that is, the number of days MUST
        be added first, followed by the number of hours, number of
        minutes, and number of seconds.

    Example:
        A duration of 15 days, 5 hours, and 20 seconds would be:

        .. code-block:: ics

            P15DT5H0M20S

        A duration of 7 weeks would be:

        .. code-block:: ics

            P7W

        .. code-block:: pycon

            >>> from icalendar.prop import vDuration
            >>> duration = vDuration.from_ical('P15DT5H0M20S')
            >>> duration
            datetime.timedelta(days=15, seconds=18020)
            >>> duration = vDuration.from_ical('P7W')
            >>> duration
            datetime.timedelta(days=49)
    """

    default_value: ClassVar[str] = "DURATION"
    params: Parameters

    def __init__(self, td: timedelta | str, /, params: dict[str, Any] | None = None):
        if isinstance(td, str):
            td = vDuration.from_ical(td)
        if not isinstance(td, timedelta):
            raise TypeError("Value MUST be a timedelta instance")
        self.td = td
        self.params = Parameters(params)

    def to_ical(self):
        sign = ""
        td = self.td
        if td.days < 0:
            sign = "-"
            td = -td
        timepart = ""
        if td.seconds:
            timepart = "T"
            hours = td.seconds // 3600
            minutes = td.seconds % 3600 // 60
            seconds = td.seconds % 60
            if hours:
                timepart += f"{hours}H"
            if minutes or (hours and seconds):
                timepart += f"{minutes}M"
            if seconds:
                timepart += f"{seconds}S"
        if td.days == 0 and timepart:
            return str(sign).encode("utf-8") + b"P" + str(timepart).encode("utf-8")
        return (
            str(sign).encode("utf-8")
            + b"P"
            + str(abs(td.days)).encode("utf-8")
            + b"D"
            + str(timepart).encode("utf-8")
        )

    @staticmethod
    def from_ical(ical):
        match = DURATION_REGEX.match(ical)
        if not match:
            raise InvalidCalendar(f"Invalid iCalendar duration: {ical}")

        sign, weeks, days, hours, minutes, seconds = match.groups()
        value = timedelta(
            weeks=int(weeks or 0),
            days=int(days or 0),
            hours=int(hours or 0),
            minutes=int(minutes or 0),
            seconds=int(seconds or 0),
        )

        if sign == "-":
            value = -value

        return value

    @property
    def dt(self) -> timedelta:
        """The time delta for compatibility."""
        return self.td

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vDuration."""
        return [cls(timedelta(1, 99))]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        return [
            name,
            self.params.to_jcal(),
            self.VALUE.lower(),
            self.to_ical().decode(),
        ]

    @classmethod
    def parse_jcal_value(cls, jcal: str) -> timedelta:
        """Parse a jCal string to a :py:class:`datetime.timedelta`.

        Raises:
            ~error.JCalParsingError: If it can't parse a duration."""
        JCalParsingError.validate_value_type(jcal, str, cls)
        try:
            return cls.from_ical(jcal)
        except (ValueError, InvalidCalendar) as e:
            raise JCalParsingError("Cannot parse duration.", cls, value=jcal) from e

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        with JCalParsingError.reraise_with_path_added(3):
            duration = cls.parse_jcal_value(jcal_property[3])
        return cls(
            duration,
            Parameters.from_jcal_property(jcal_property),
        )


__all__ = ["vDuration"]



"""Parsing a component's iCalendar data."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from icalendar.parser.content_line import Contentline, Contentlines
from icalendar.parser.property import split_on_unescaped_comma
from icalendar.prop import vBroken
from icalendar.timezone import tzp

if TYPE_CHECKING:
    from icalendar.cal import Component, ComponentFactory
    from icalendar.parser.parameter import Parameters
    from icalendar.prop import VPROPERTY, TypesFactory


class ComponentIcalParser:
    """A parser for a component's iCalendar data.

    This uses the template method pattern, where the main parsing
    logic can be refined in subclasses.
    """

    datetime_names: ClassVar[tuple[str, ...]] = (
        "DTSTART",
        "DTEND",
        "RECURRENCE-ID",
        "DUE",
        "RDATE",
        "EXDATE",
    )
    """Names to check for TZID parameter when parsing datetimes.

    Their ``from_ical`` methods take an optional ``tzid`` argument,
    which is used if the property has a TZID parameter.
    """

    def __init__(
        self,
        data: bytes | str | list[Contentline],
        component_factory: ComponentFactory,
        types_factory: TypesFactory,
    ):
        """Initialize the parser with the raw data.

        Parameters:
            data: The raw iCalendar data to parse, either as bytes
                or a list of content lines.
            component_factory: The factory to use for creating components.
            types_factory: The factory to use for creating property values.
        """
        self._data = data
        self._component_factory = component_factory
        self._types_factory = types_factory
        self._tzp = tzp

    _content_lines: list[Contentline]

    def contains_component(self, name: str) -> bool:
        """Check if the parser contains a component."""
        self.initialize_parsing()
        begin_line = "BEGIN:" + name.upper()
        return any(
            len(content_line) == len(begin_line) and content_line.upper() == begin_line
            for content_line in self._content_lines
        )

    def contains_uid(self, uid: str) -> bool:
        """Determines whether the component contains a ``uid``.

        Returns:
            ``True`` if the component contains a ``uid``, else ``False``.
        """
        self.initialize_parsing()
        return any(uid in line for line in self._content_lines)

    def initialize_parsing(self):
        self._stack: list[Component] = []
        self._components: list[Component] = []
        self._data = self._content_lines = (
            self._data
            if isinstance(self._data, list)
            else Contentlines.from_ical(self._data)
        )
        self._content_lines_iterator = iter(self._content_lines)

    def handle_line_parse_error(self, exception: Exception):
        """Handle a line parsing error."""
        # if unable to parse a line within a component
        # that ignores exceptions, mark the component
        # as broken and skip the line. otherwise raise.
        component = self.component
        if not component or not component.ignore_exceptions:
            raise exception
        component.errors.append((None, str(exception)))

    def handle_begin_component(self, vals: str) -> None:
        """Handle the beginning of a component."""
        # try and create one of the components defined in the spec,
        # otherwise get a general Components for robustness.
        c_name = vals.upper()
        c_class = self._component_factory.get_component_class(c_name)
        # If component factory cannot resolve ``c_name``, the generic
        # ``Component`` class is used which does not have the name set.
        # That's opposed to the usage of ``cls``, which represents a
        # more concrete subclass with a name set (e.g. VCALENDAR).
        component = c_class()
        if not getattr(component, "name", ""):  # undefined components
            component.name = c_name
        self._stack.append(component)

    def handle_end_component(self, vals: str) -> None:
        """Handle the end of a component."""
        # we are done adding properties to this component
        # so pop it from the stack and add it to the new top.
        if not self._stack:
            # The stack is currently empty, the input must be invalid
            raise ValueError("END encountered without an accompanying BEGIN!")

        component = self._stack.pop()
        if not self._stack:  # we are at the end
            self._components.append(component)
        else:
            self._stack[-1].add_component(component)
        if vals.upper() == "VTIMEZONE" and "TZID" in component:
            tzp.cache_timezone_component(component)

    def prepare_components(self) -> None:
        """Prepare the parsed components.

        This is called when all components are parsed.
        """

    def parse(self) -> list[Component]:
        """Parse the raw data."""
        self.initialize_parsing()
        self.parse_content_lines()
        self.prepare_components()
        return self._components

    def parse_content_lines(self) -> None:
        """Parse the content lines."""
        for line in self._content_lines_iterator:
            if not line:
                continue
            try:
                name, params, vals = line.parts()
            except ValueError as e:
                self.handle_line_parse_error(e)
                continue

            uname = name.upper()
            if uname == "BEGIN":
                self.handle_begin_component(vals)
            elif uname == "END":
                self.handle_end_component(vals)
            else:
                self.handle_property(uname, params, vals, line)

    @property
    def component(self) -> Component | None:
        return self._stack[-1] if self._stack else None

    def get_factory_for_property(self, name: str, params: Parameters) -> VPROPERTY:
        """Get the factory for a property."""
        return self._types_factory.for_property(name, params.value)

    def handle_property(
        self, name: str, params: Parameters, vals: str, line: Contentline
    ) -> None:
        """Handle a property line.

        Add properties to the top of the current stack.

        Parameters:
            name: The name of the property, uppercased.
            params: The parameters of the property.
            vals: The value of the property.
            line: The original content line.
        """
        # Extract VALUE parameter if present
        if not self.component:
            # only accept X-COMMENT at the end of the .ics file
            # ignore these components in parsing
            if name == "X-COMMENT":
                return
            raise ValueError(f'Property "{name}" does not have a parent component.')
        # Determine TZID for datetime properties
        tzid = params.get("TZID") if params and name in self.datetime_names else None

        # Handle special cases for value list preparation
        if name == "CATEGORIES":
            if self.handle_categories(params, vals, line):
                return
            # Fallback to normal processing if we can't find colon
            vals_list = [vals]
        elif name == "FREEBUSY":
            # Handle FREEBUSY comma-separated values
            vals_list = vals.split(",")
        # Workaround broken ICS files with empty RDATE
        # (not EXDATE - let it parse and fail)
        elif name == "RDATE" and vals == "":
            vals_list = []
        else:
            vals_list = [vals]

        # Parse all properties eagerly
        for val in vals_list:
            self.parse_and_add_property(name, params, val, tzid, line)

    def parse_and_add_property(
        self,
        name: str,
        params: Parameters,
        val: str,
        tzid: str | None,
        line: Contentline,
    ):
        """Parse a property value and add it to the current component."""
        factory = self.get_factory_for_property(name, params)
        try:
            if tzid:
                parsed_val = factory.from_ical(val, tzid)
            else:
                parsed_val = factory.from_ical(val)
        except (ValueError, TypeError) as e:
            self.handle_property_parse_error(e, name, params, val, line)
        else:
            vals_inst = factory(parsed_val)
            vals_inst.params = params
            self.component.add(name, vals_inst, encode=False)

    def handle_property_parse_error(
        self,
        exception: Exception,
        name: str,
        params: Parameters,
        val: str,
        line: Contentline,
    ):
        """Handle the parse error for a property."""
        if not self.component.ignore_exceptions and not name[:2].upper() == "X-":
            raise exception
        # Error-tolerant mode: create vBroken
        factory = self.get_factory_for_property(name, params)
        expected_type = getattr(factory, "__name__", "unknown")
        broken_prop = vBroken.from_parse_error(
            raw_value=val,
            params=params,
            property_name=name,
            expected_type=expected_type,
            error=exception,
        )
        self.component.errors.append((name, str(exception)))
        self.component.add(name, broken_prop, encode=0)

    def handle_categories(
        self, params: Parameters, vals: str, line: Contentline
    ) -> bool:
        """Handle the special case of CATEGORIES property.

        Returns:
            ``True`` if handled, else ``False``.
        """
        # Special handling for CATEGORIES - need raw value
        # before unescaping to properly split on unescaped commas
        line_str = str(line)
        # Use rfind to get the last colon (value separator)
        # to handle parameters with colons like ALTREP="http://..."
        colon_idx = line_str.rfind(":")
        if colon_idx > 0:
            raw_value = line_str[colon_idx + 1 :]
            # Parse categories immediately (not lazily) for both
            # strict and tolerant components.
            # CATEGORIES needs special comma handling
            try:
                category_list = split_on_unescaped_comma(raw_value)
                factory = self.get_factory_for_property("CATEGORIES", params)
                vals_inst = factory(category_list)
                vals_inst.params = params
                self.component.add("CATEGORIES", vals_inst, encode=0)
            except ValueError as e:
                self.handle_property_parse_error(
                    e, "CATEGORIES", params, raw_value, line
                )
            return True
        return False



"""Tools for parsing properties."""

import re

from icalendar.parser.string import _unescape_string


def unescape_list_or_string(val: str | list[str]) -> str | list[str]:
    """Unescape a value that may be a string or list of strings.

    Applies :func:`_unescape_string` to the value. If the value is a list,
    unescapes each element.

    Parameters:
        val: A string or list of strings to unescape.

    Returns:
        The unescaped values.
    """
    if isinstance(val, list):
        return [_unescape_string(s) for s in val]
    return _unescape_string(val)


_unescape_backslash_regex = re.compile(r"\\([\\,;:nN])")


def unescape_backslash(val: str):
    r"""Unescape backslash sequences in iCalendar text.

    Unlike :py:meth:`unescape_string`, this only handles actual backslash escapes
    per :rfc:`5545`, not URL encoding. This preserves URL-encoded values
    like ``%3A`` in URLs.

    Processes backslash escape sequences in a single pass using regex matching.
    """
    return _unescape_backslash_regex.sub(
        lambda m: "\n" if m.group(1) in "nN" else m.group(1), val
    )


def split_on_unescaped_comma(text: str) -> list[str]:
    r"""Split text on unescaped commas and unescape each part.

    Splits only on commas not preceded by backslash.
    After splitting, unescapes backslash sequences in each part.

    Parameters:
        text: Text with potential escaped commas (e.g., "foo\\, bar,baz")

    Returns:
        List of unescaped category strings

    Examples:
        .. code-block:: pycon

            >>> from icalendar.parser import split_on_unescaped_comma
            >>> split_on_unescaped_comma(r"foo\, bar,baz")
            ['foo, bar', 'baz']
            >>> split_on_unescaped_comma("a,b,c")
            ['a', 'b', 'c']
            >>> split_on_unescaped_comma(r"a\,b\,c")
            ['a,b,c']
            >>> split_on_unescaped_comma(r"Work,Personal\,Urgent")
            ['Work', 'Personal,Urgent']
    """
    if not text:
        return [""]

    result = []
    current = []
    i = 0

    while i < len(text):
        if text[i] == "\\" and i + 1 < len(text):
            # Escaped character - keep both backslash and next char
            current.append(text[i])
            current.append(text[i + 1])
            i += 2
        elif text[i] == ",":
            # Unescaped comma - split point
            result.append(unescape_backslash("".join(current)))
            current = []
            i += 1
        else:
            current.append(text[i])
            i += 1

    # Add final part
    result.append(unescape_backslash("".join(current)))

    return result


def split_on_unescaped_semicolon(text: str) -> list[str]:
    r"""Split text on unescaped semicolons and unescape each part.

    Splits only on semicolons not preceded by a backslash.
    After splitting, unescapes backslash sequences in each part.
    Used by vCard structured properties (ADR, N, ORG) per :rfc:`6350`.

    Parameters:
        text: Text with potential escaped semicolons (e.g., "field1\\;with;field2")

    Returns:
        List of unescaped field strings

    Examples:
        .. code-block:: pycon

            >>> from icalendar.parser import split_on_unescaped_semicolon
            >>> split_on_unescaped_semicolon(r"field1\;with;field2")
            ['field1;with', 'field2']
            >>> split_on_unescaped_semicolon("a;b;c")
            ['a', 'b', 'c']
            >>> split_on_unescaped_semicolon(r"a\;b\;c")
            ['a;b;c']
            >>> split_on_unescaped_semicolon(r"PO Box 123\;Suite 200;City")
            ['PO Box 123;Suite 200', 'City']
    """
    if not text:
        return [""]

    result = []
    current = []
    i = 0

    while i < len(text):
        if text[i] == "\\" and i + 1 < len(text):
            # Escaped character - keep both backslash and next char
            current.append(text[i])
            current.append(text[i + 1])
            i += 2
        elif text[i] == ";":
            # Unescaped semicolon - split point
            result.append(unescape_backslash("".join(current)))
            current = []
            i += 1
        else:
            current.append(text[i])
            i += 1

    # Add final part
    result.append(unescape_backslash("".join(current)))

    return result


__all__ = [
    "split_on_unescaped_comma",
    "split_on_unescaped_semicolon",
    "unescape_backslash",
    "unescape_list_or_string",
]



"""CATEGORIES property values from :rfc:`5545`."""

from collections.abc import Iterator
from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.parser_tools import to_unicode
from icalendar.prop.text import vText


class vCategory:
    default_value: ClassVar[str] = "TEXT"
    params: Parameters

    def __init__(
        self, c_list: list[str] | str, /, params: dict[str, Any] | None = None
    ):
        if not hasattr(c_list, "__iter__") or isinstance(c_list, str):
            c_list = [c_list]
        self.cats: list[vText | str] = [vText(c) for c in c_list]
        self.params = Parameters(params)

    def __iter__(self) -> Iterator[vText | str]:
        return iter(self.cats)

    def to_ical(self) -> bytes:
        return b",".join(
            [
                c.to_ical() if hasattr(c, "to_ical") else vText(c).to_ical()
                for c in self.cats
            ]
        )

    @staticmethod
    def from_ical(ical: list[str] | str) -> list[str]:
        """Parse a CATEGORIES value from iCalendar format.

        This helper is normally called by :meth:`Component.from_ical`, which
        already splits the CATEGORIES property into a list of unescaped
        category strings. New code should therefore pass a list of strings.

        Passing a single comma-separated string is supported only for
        backwards compatibility with older parsing code and is considered
        legacy behavior.

        Parameters:
            ical: A list of category strings (preferred, as provided by
                :meth:`Component.from_ical`), or a single comma-separated
                string from a legacy caller.

        Returns:
            A list of category strings.
        """
        if isinstance(ical, list):
            # Already split by Component.from_ical()
            return ical
        # Legacy: simple comma split (no escaping handled)
        ical = to_unicode(ical)
        return ical.split(",")

    def __eq__(self, other: object) -> bool:
        """self == other"""
        return isinstance(other, vCategory) and self.cats == other.cats

    def __hash__(self) -> int:
        """Hash of the vCategory object."""
        return hash(tuple(self.cats))

    def __repr__(self) -> str:
        """String representation."""
        return f"{self.__class__.__name__}({self.cats}, params={self.params})"

    def to_jcal(self, name: str) -> list:
        """The jCal representation for categories."""
        result = [name, self.params.to_jcal(), self.VALUE.lower()]
        result.extend(map(str, self.cats))
        if not self.cats:
            result.append("")
        return result

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vCategory."""
        return [cls(["HOME", "COSY"])]

    from icalendar.param import VALUE

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        for i, category in enumerate(jcal_property[3:], start=3):
            JCalParsingError.validate_value_type(category, str, cls, i)
        return cls(
            jcal_property[3:],
            Parameters.from_jcal_property(jcal_property),
        )

    @property
    def ical_value(self) -> list[str]:
        """The list of categories as strings."""
        return [str(cat) for cat in self.cats]


__all__ = ["vCategory"]



"""Components for lazy parsing of components."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from icalendar.cal.component_factory import ComponentFactory
from icalendar.parser.ical.lazy import LazyCalendarIcalParser

from .calendar import Calendar

if TYPE_CHECKING:
    from collections.abc import Callable

    from icalendar.parser.ical.component import ComponentIcalParser
    from icalendar.parser.ical.lazy import LazySubcomponent

    from .component import Component


class ParsedSubcomponentsStrategy:
    """All the subcomponents are parsed and available as a list."""

    def __init__(self):
        self._components: list[Component] = []

    def get_all_components(self) -> tuple[ParsedSubcomponentsStrategy, list[Component]]:
        """Get the subcomponents of the calendar."""
        return self, self._components

    def set_components(
        self, components: list[Component]
    ) -> ParsedSubcomponentsStrategy:
        """Set the subcomponents of the calendar."""
        self._components = components
        return self

    def add_component(self, component: Component) -> ParsedSubcomponentsStrategy:
        """Add a component to the calendar."""
        self._components.append(component.parse())
        return self

    def is_lazy(self) -> Literal[False]:
        """Returns ``False`` because subcomponents are not lazily parsed."""
        return False

    def walk(self, name: str) -> tuple[ParsedSubcomponentsStrategy, list[Component]]:
        """Get the subcomponents of the calendar with the given name."""
        result = []
        for component in self._components:
            result += component.walk(name)
        return self, result

    def with_uid(
        self, name: str
    ) -> tuple[ParsedSubcomponentsStrategy, list[Component]]:
        """Get the subcomponents of the calendar with the given uid."""
        result = []
        for component in self._components:
            result += component.with_uid(name)
        return self, result


class LazySubcomponentsStrategy:
    """Parse subcomponents only when accessed."""

    initial_components_to_parse: tuple[str, ...] = ("VTIMEZONE",)
    """Parse these subcomponents before any others."""

    def __init__(self):
        self._components: list[LazySubcomponent | Component] = []
        self._initial_parsed: bool = False

    @property
    def as_parsed(self) -> ParsedSubcomponentsStrategy:
        """Return the parsed components."""
        return ParsedSubcomponentsStrategy().set_components(
            [component.parse() for component in self._components]
        )

    def get_all_components(self) -> tuple[ParsedSubcomponentsStrategy, list[Component]]:
        """Get the subcomponents of the calendar.

        Parse all subcomponents.
        """
        self.parse_initial_components()
        return self.as_parsed.get_all_components()

    def set_components(
        self, components: list[Component]
    ) -> ParsedSubcomponentsStrategy:
        """Set the subcomponents of the calendar."""
        return ParsedSubcomponentsStrategy().set_components(components)

    def add_component(
        self, component: Component | LazySubcomponent
    ) -> LazySubcomponentsStrategy:
        """Add a component to the calendar."""
        self._components.append(component)
        return self

    def is_lazy(self) -> bool:
        """Return whether the subcomponents may be lazily parsed."""
        return True

    def parse_initial_components(self) -> None:
        """Parse the components that are required by other components.

        This mainly concerns the timezone components.
        They are required by other components that have a TZID parameter.
        """
        if self._initial_parsed:
            return
        self._initial_parsed = True
        for component in self._components:
            if component.name in self.initial_components_to_parse:
                component.parse()

    def walk(
        self, name: str | None
    ) -> tuple[LazySubcomponentsStrategy, list[Component]]:
        """Get the subcomponents of the calendar with the given name.

        Parse only the minimal number of subcomponents.
        """
        if name is None:
            return self.as_parsed.walk(name)
        self.parse_initial_components()
        result = []
        for component in self._components:
            result += component.walk(name)
        return self, result

    def with_uid(self, uid: str) -> tuple[LazySubcomponentsStrategy, list[Component]]:
        """Get the subcomponents of the calendar with the given ``uid``.

        Parse only the minimal number of subcomponents.
        """
        self.parse_initial_components()
        result = []
        for component in self._components:
            result += component.with_uid(uid)
        return self, result


class InitialSubcomponentsStrategy:
    """Initial strategy for the calendar.

    No subcomponents.
    """

    def set_components(self, components: list[Component]) -> LazySubcomponentsStrategy:
        if components:
            raise ValueError(
                "Cannot set subcomponents on an uninitialised LazyCalendar. "
                "Parse it first or add components via add_component()."
            )
        return LazySubcomponentsStrategy()


class LazyCalendar(Calendar):
    """A calendar that can handle big files.

    Subcomponents of this calendar are evaluated lazily,
    meaning that they are not parsed until they are accessed.
    This allows the calendar to handle large files without
    consuming too much memory or time.

    All properties of the calendar component are parsed immediately.
    Subcomponents and their properties are parsed lazily.

    Examples:

        By accessing the :attr:`~icalendar.cal.calendar.Calendar.events` of the calendar,
        only :class:`~icalendar.cal.event.Event` and
        :class:`~icalendar.cal.timezone.Timezone` are immediately parsed.

        .. code-block:: pycon

            >>> from icalendar import LazyCalendar
            >>> calendar = LazyCalendar.example("issue_1050_all_components")
            >>> len(calendar.events) == 1
            True

        The calendar's subcomponents were not parsed because they were not accessed.
        The calendar is still lazy.

            >>> calendar.is_lazy()
            True

        When you access all :attr:`subcomponents` of the calendar,
        for example by getting their count, the entire calendar is
        parsed and becomes not lazy.

            >>> len(calendar.subcomponents)
            5
            >>> calendar.is_lazy()
            False

    """

    _subcomponents: (
        LazySubcomponentsStrategy
        | ParsedSubcomponentsStrategy
        | InitialSubcomponentsStrategy
    )
    """The strategy pattern for subcomponents of the calendar."""

    def __init__(self, *args, **kwargs):
        """Initialize the calendar."""
        self._subcomponents = InitialSubcomponentsStrategy()
        super().__init__(*args, **kwargs)

    @property
    def subcomponents(self) -> list[Component]:
        """The subcomponents of the calendar.

        Parse all subcomponents of the calendar and return them as a list.

        You can manipulate this list or set it.
        It has the same behavior as in :class:`~icalendar.cal.calendar.Calendar`.
        """
        self._subcomponents, result = self._subcomponents.get_all_components()
        return result

    @subcomponents.setter
    def subcomponents(self, value: list[Component]) -> None:
        """Set the subcomponents of the calendar."""
        self._subcomponents = self._subcomponents.set_components(value)

    @classmethod
    def _get_ical_parser(cls, st: str | bytes) -> ComponentIcalParser:
        """Get the iCal parser for the given input string."""
        return LazyCalendarIcalParser(
            st, cls._get_component_factory(), cls.types_factory
        )

    @classmethod
    def _get_component_factory(cls) -> ComponentFactory:
        """Get the component factory for this calendar."""
        factory = ComponentFactory()
        factory.add_component_class(cls)
        return factory

    def add_component(self, component: Component) -> None:
        """Add a component to the calendar.

        Use this instead of appending to
        :attr:`~icalendar.cal.lazy.LazyCalendar.subcomponents`,
        as the latter does not parse the whole calendar.
        """
        self._subcomponents = self._subcomponents.add_component(component)

    def is_lazy(self) -> bool:
        """Whether the subcomponents will be parsed lazily.

        .. note:: If you believe that the calendar parses more than it should,
            please `open an issue <https://github.com/collective/icalendar/issues/new?template=bug_report.md>`_.

        Returns:
            ``True`` if subcomponents are deferred and not yet parsed.
            ``False`` if all subcomponents have been parsed.
        """
        return self._subcomponents.is_lazy()

    def _walk(
        self, name: str | None, select: Callable[[Component], bool]
    ) -> list[Component]:
        self._subcomponents, result = self._subcomponents.walk(name)
        result = [component for component in result if select(component)]
        if (name is None or self.name == name) and select(self):
            result.insert(0, self)
        return result

    def with_uid(self, uid: str) -> list[Component]:
        self._subcomponents, result = self._subcomponents.with_uid(uid)
        if self.uid == uid:
            result.insert(0, self)
        return result

    with_uid.__doc__ = Calendar.with_uid.__doc__


__all__ = ["LazyCalendar"]

if __name__ == "__main__":
    import timeit

    calendar = Calendar.example("issue_1050_all_components")
    COUNT = 10000
    calendar.subcomponents *= COUNT
    ics = calendar.to_ical()

    def _benchmark(cal: type[Calendar]):
        """Check out how fast this is."""
        cal = cal.from_ical(ics)
        assert len(cal.events) == COUNT

    for cal in [Calendar, LazyCalendar]:
        print("Benchmarking:", cal.__name__)  # noqa: T201
        print(timeit.timeit("_benchmark(cal)", globals=locals(), number=1))  # noqa: T201

    # Benchmarking: Calendar
    # 12.277852076000272
    # Benchmarking: LazyCalendar
    # 5.738950790999297
